import numpy as np
from nleis.nleis import EISandNLEIS
import os

import os.path
import sys


def test_model_io():

    test_dir = os.path.dirname(os.path.abspath(__file__))
    data_dir = os.path.join(test_dir, '../data')

    # # get example data
    # # The example is shown in "Getting Started" page 
    
    frequencies = np.loadtxt(os.path.join(data_dir, 'freq_30a.txt'))
    Z1 = np.loadtxt(os.path.join(data_dir,'Z1s_30a.txt')).view(complex)[1]
    Z2 = np.loadtxt(os.path.join(data_dir,'Z2s_30a.txt')).view(complex)[1]
    circ_str_1 = 'L0-R0-TDS0-TDS1'
    circ_str_2 = 'd(TDSn0,TDSn1)'

    initial_guess = [1e-7,1e-3 # L0,RO
                    ,5e-3,1e-3,10,1e-2,100,10,0.1 ## TDS0 + additioal nonlinear parameters
                    ,1e-3,1e-3,1e-3,1e-2,1000,0,0 ## TDS1 + additioal nonlinear parameters
                    ]

    circuit_1 = EISandNLEIS(circ_str_1,circ_str_2,initial_guess=initial_guess)

    # circuit_1.save('./nleis_tests/test_io.json')
    circuit_1.save(os.path.join(data_dir,'test_io.json'))

    circuit_2 = EISandNLEIS()
    circuit_2.load(os.path.join(data_dir,'test_io.json'))

    assert circuit_1 == circuit_2

    circuit_1.fit(frequencies, Z1,Z2)
    p_fit = list(circuit_1.parameters_)
    circuit_1.save(os.path.join(data_dir,'test_io.json'))
    circuit_2 = EISandNLEIS()
    circuit_2.load(os.path.join(data_dir,'test_io.json'))

    assert str(circuit_1) == str(circuit_2)
    assert circuit_1 == circuit_2

    fitted_template = EISandNLEIS()
    fitted_template.load(os.path.join(data_dir,'test_io.json'), fitted_as_initial=True)
    circuit_1 = EISandNLEIS(circ_str_1,circ_str_2,initial_guess=p_fit)
    assert str(circuit_1) == str(fitted_template)
    assert circuit_1 == fitted_template



import json
import os

import numpy as np
import matplotlib.pyplot as plt
import pytest

from impedance.models.circuits import BaseCircuit, CustomCircuit, Randles
from nleis.nleis import EISandNLEIS, NLEISCustomCircuit

test_dir = os.path.dirname(os.path.abspath(__file__))
data_dir = os.path.join(test_dir, '../data')

# # get example data
# # The example is shown in "Getting Started" page 

frequencies = np.loadtxt(os.path.join(data_dir, 'freq_30a.txt'))
Z1 = np.loadtxt(os.path.join(data_dir,'Z1s_30a.txt')).view(complex)[1]
Z2 = np.loadtxt(os.path.join(data_dir,'Z2s_30a.txt')).view(complex)[1]




def test_EISandNLEIS():
    circ_str_1 = 'L0-R0-TDS0-TDS1'
    circ_str_2 = 'd(TDSn0,TDSn1)'
    initial_guess = [1e-7,1e-3 # L0,RO
                    ,5e-3,1e-3,10,1e-2,100,10,0.1 ## TDS0 + additioal nonlinear parameters
                    ,1e-3,1e-3,1e-3,1e-2,1000,0,0 ## TDS1 + additioal nonlinear parameters
                    ]

    NLEIS_circuit = EISandNLEIS(circ_str_1,circ_str_2,initial_guess=initial_guess)

    # check get_param_names()
    full_names_EIS, all_units_EIS = NLEIS_circuit.get_param_names(circ_str_1,{})
    assert full_names_EIS == ['L0', 'R0', 'TDS0_0', 'TDS0_1','TDS0_2','TDS0_3','TDS0_4',
                              'TDS1_0','TDS1_1','TDS1_2','TDS1_3','TDS1_4']
    assert all_units_EIS == ['H','Ohm','Ohms', 'Ohms', 'F','Ohms','s','Ohms', 'Ohms', 'F','Ohms','s']
    full_names_NLEIS, all_units_NLEIS = NLEIS_circuit.get_param_names(circ_str_2,{})
    assert full_names_NLEIS == ['TDSn0_0', 'TDSn0_1','TDSn0_2','TDSn0_3','TDSn0_4','TDSn0_5','TDSn0_6'
                                ,'TDSn1_0','TDSn1_1','TDSn1_2','TDSn1_3','TDSn1_4','TDSn1_5','TDSn1_6']
    assert all_units_NLEIS == ['Ohms', 'Ohms', 'F','Ohms','s','1/V','','Ohms', 'Ohms', 'F','Ohms','s','1/V','']

    # check _is_fit()
    assert not NLEIS_circuit._is_fit()

    # check complex frequencies raise TypeError
    with pytest.raises(TypeError):
        NLEIS_circuit.predict([0.42, 42 + 42j])





from .nleis_fitting import simul_fit,wrappedImpedance,individual_parameters
from .nleis_elements_pair import*
from impedance.models.circuits.fitting import check_and_eval
from impedance.models.circuits.elements import circuit_elements,get_element_from_name
from impedance.models.circuits.circuits import BaseCircuit
from .fitting import circuit_fit, buildCircuit,calculateCircuitLength,extract_circuit_elements

from impedance.visualization import plot_bode
from .visualization import plot_altair, plot_first, plot_second

import json
import matplotlib.pyplot as plt
import numpy as np
import warnings



class EISandNLEIS:
    ### ToDO: add SSO method and SO method 
    def __init__(self, circuit_1='',circuit_2='',initial_guess = []
                 ,constants = None, name=None, **kwargs):
        """ 
        
        Constructor for a customizable equivalent circuit model

        Parameters
        ----------

        circuit_1: string
            A string that should be interpreted as an equivalent circuit for linear EIS
            
        circuit_2: string
            A string that should be interpreted as an equivalent circuit for NLEIS
            
        initial_guess: numpy array
                Initial guess of the circuit values
                
        constants : dict, optional
            Parameters and values to hold constant during fitting
            (e.g. {"R0": 0.1})

        name: str, optional
            Name for the model

        Notes
        -----
        A custom circuit is defined as a string comprised of elements in series
        (separated by a `-`) and elements in parallel (grouped as (x,y)) for EIS.
        For NLEIS, the circuit should be grouped by d(cathode,anode)
        Each element can be appended with an integer (e.g. R0) or an underscore
        and an integer (e.g. CPE_1) to make keeping track of multiple elements
        of the same type easier.

        Example:
            A two electrode cell with sperical porous cathode and anode, resistor, and inductor is represents as,
            EIS: circuit_1 = 'L0-R0-TDS0-TDS1'
            NLEIS: circuit_2 = 'd(TDSn0-TDSn1)'

        """
        # if supplied, check that initial_guess is valid and store
        initial_guess = [x for x in initial_guess if x is not None]
        for i in initial_guess:
            if not isinstance(i, (float, int, np.int32, np.float64)):
                raise TypeError(f'value {i} in initial_guess is not a number')

        # initalize class attributes
        # self.initial_guess = list(initial_guess)
        self.initial_guess = initial_guess

        self.circuit_1 = circuit_1
        self.circuit_2 = circuit_2
        elements_1 = extract_circuit_elements(circuit_1)
        elements_2 = extract_circuit_elements(circuit_2)
        for element in elements_2:
            if element.replace('n', '') not in elements_1 or ('n' not in element and element != ''):
                raise TypeError('The pairing of linear and nonlinear elements must be presented correctly for simultaneous fitting.'+ ' Double check the EIS circuit: ' + f'{circuit_1}' + ' and the NLEIS circuit: '+ f'{circuit_2}' +'. The parsed elements are '+f'{elements_1}'+' for circuit_1, and ' + f'{elements_2}'+ ' for circuit_2.')
                
            
        input_elements = elements_1 + elements_2
             
        if constants is not None:
            self.constants = constants
            self.constants_1 = {}
            self.constants_2 = {}
            ## New code for constant separatation
            ## in the current development the differentiation between linear and nonlinear is separated by 'n'
            ## using get_element_from_name to get the raw element and adding n to the end
            for elem in self.constants:
                
                raw_elem = get_element_from_name(elem)
                raw_circuit_elem = elem.split('_')[0]
                
                if raw_circuit_elem not in input_elements:
                    raise ValueError(f'{raw_elem} not in ' +
                         f'input elements ({input_elements})')
                raw_num_params = check_and_eval(raw_elem).num_params
                if raw_num_params<=1:
                    ## currently there is no single parameter nonlinear element,
                    ## so this can work, but might be changed in the future
                    self.constants_1[elem]=self.constants[elem]
                else:
                    param_num = int(elem.split('_')[-1])
                    if raw_elem[-1] != 'n':
                        
                        if param_num>=raw_num_params:
                            raise ValueError(f'{elem} is out of the range of the maximum allowed ' +
                                             f'number of parameters ({raw_num_params})')
    
                        self.constants_1[elem]=self.constants[elem]
                        len_elem = len(raw_elem)
                        nl_elem = elem[0:len_elem]+'n'+elem[len_elem:]
                        raw_nl_elem = get_element_from_name(nl_elem)
                        if raw_nl_elem in circuit_elements.keys():
                            self.constants_2[nl_elem]=self.constants[elem]
                        else:
                            self.constants_2[elem]=self.constants[elem]
    
                    if raw_elem[-1] == 'n':
                        
                        if param_num>=raw_num_params:
                            raise ValueError(f'{elem} is out of the range of the maximum allowed ' +
                                             f'number of parameters ({raw_num_params})')
                        
                        num_params = check_and_eval(raw_elem[0:-1]).num_params
                        len_elem = len(raw_elem[0:-1])
                        if param_num<num_params:
                            self.constants_1[elem[0:len_elem]+elem[len_elem+1:]]=self.constants[elem]
                        self.constants_2[elem]=self.constants[elem]

        else:
            self.constants = {}
            self.constants_1 = {}
            self.constants_2 = {}
        ### new code for circuit length calculation using circuit element
        ### producing edited circuit 
        ### i.e. circuit_1 = L0-R0-TDP0-TDS1; circuit_2 = TDPn0-TDSn1; edited_circuit = L0-R0-TDPn0-TDSn1
        if elements_1 != [''] or elements_2 !=['']:
            if elements_1 == [''] or elements_2 ==['']:
                raise ValueError('Either circuit_1 or circuit_2 cannot be empty')
            edited_circuit = ''
            for elem in elements_1:
                nl_elem = elem[0:-1]+'n'+elem[-1]
                if nl_elem in elements_2:
                    edited_circuit += '-'+ nl_elem
                else:
                    edited_circuit += '-'+ elem
            self.edited_circuit = edited_circuit[1:]
        else:
            self.edited_circuit = ''
            
        circuit_len = calculateCircuitLength(self.edited_circuit)      

        if len(self.initial_guess) + len(self.constants) != circuit_len:
            raise ValueError('The number of initial guesses ' +
                             f'({len(self.initial_guess)}) + ' +
                             'the number of constants ' +
                             f'({len(self.constants)})' +
                             ' must be equal to ' +
                             f'the circuit length ({circuit_len})')
        self.name = name

        # initialize fit parameters and confidence intervals
        self.parameters_ = None
        self.conf_ = None

        self.p1, self.p2 = individual_parameters(self.circuit_1,self.initial_guess,self.constants_1,self.constants_2)

    def __eq__(self, other):
        if self.__class__ == other.__class__:
            matches = []
            for key, value in self.__dict__.items():
                if isinstance(value, np.ndarray):
                    matches.append((value == other.__dict__[key]).all())
                else:
                    matches.append(value == other.__dict__[key])
            return np.array(matches).all()
        else:
            raise TypeError('Comparing object is not of the same type.')
        
    def fit(self, frequencies, Z1,Z2, bounds = None,
            opt='max',max_f=10,cost = 0.5, **kwargs):
        """ 
        
        Fit the circuit model

        Parameters
        ----------
        frequencies: numpy array
            Frequencies

        Z1: numpy array of dtype 'complex128'
            EIS values to fit
        Z1: numpy array of dtype 'complex128'
            NLEIS values to fit

        bounds: 2-tuple of array_like, optional
            Lower and upper bounds on parameters. When bounds are provided, the input will normalized to stabilize the algorithm

        Normalization : str, optional
            Default is max normalization. Other normalization will be supported
            in the future 
        max_f: int
            The the maximum frequency of interest for NLEIS

        kwargs :
            Keyword arguments passed to simul_fit,
            and subsequently to scipy.optimize.curve_fit
            or scipy.optimize.basinhopping

        Returns
        -------
        self: returns an instance of self

        """
        # check that inputs are valid:
        #    frequencies: array of numbers
        #    impedance: array of complex numbers
        #    impedance and frequency match in length
        
        frequencies = np.array(frequencies, dtype=float)
        Z1 = np.array(Z1, dtype=complex)
        Z2 = np.array(Z2, dtype=complex)
        if len(frequencies) != len(Z1):
            raise TypeError('length of frequencies and impedance do not match for EIS')

        if self.initial_guess != []:
            parameters, conf = simul_fit(frequencies, Z1,Z2,
                                           self.circuit_1,self.circuit_2,self.edited_circuit, self.initial_guess,
                                           constants_1=self.constants_1,constants_2=self.constants_2,
                                           bounds=bounds,
                                           opt=opt,cost = cost,max_f = max_f,
                                           **kwargs)
            # self.parameters_ = list(parameters)
            self.parameters_ = parameters

            if conf is not None:
                # self.conf_ = list(conf)
                self.conf_ = conf
                self.conf1, self.conf2 = individual_parameters(self.circuit_1,self.conf_,self.constants_1,self.constants_2)

            self.p1, self.p2 = individual_parameters(self.circuit_1,self.parameters_,self.constants_1,self.constants_2)
        else:
            raise ValueError('no initial guess supplied')

        return self

    def _is_fit(self):
        """ 
        
        check if model has been fit (parameters_ is not None) 
        
        """
        if self.parameters_ is not None:
            return True
        else:
            return False


    def predict(self, frequencies, max_f=10, use_initial=False):
        """ 
        
        Predict impedance using an equivalent circuit model

        Parameters
        ----------
        frequencies: ndarray of numeric dtype
        
        max_f: int
            The the maximum frequency of interest for NLEIS

        use_initial: boolean
            If true and the model was previously fit use the initial
            parameters instead

        Returns
        -------
        impedance: ndarray of dtype 'complex128'
            Predicted impedance
        
        """
        if not isinstance(frequencies, np.ndarray):
            raise TypeError('frequencies is not of type np.ndarray')
        if not (np.issubdtype(frequencies.dtype, np.integer) or
                np.issubdtype(frequencies.dtype, np.floating)):
            raise TypeError('frequencies array should have a numeric ' +
                            f'dtype (currently {frequencies.dtype})')

        f1 =frequencies
        mask = np.array(frequencies)<max_f
        f2 = frequencies[mask]

        if self._is_fit() and not use_initial:
            x1,x2 = wrappedImpedance(self.circuit_1, self.constants_1,self.circuit_2,self.constants_2,f1,f2,self.parameters_)   
            return x1,x2
        else:
            warnings.warn("Simulating circuit based on initial parameters")
            x1,x2 = wrappedImpedance(self.circuit_1, self.constants_1,self.circuit_2,self.constants_2,f1,f2,self.initial_guess)
            return(x1,x2)
    def get_param_names(self,circuit,constants):
        """ 
        
        Converts circuit string to names and units 
        
        """

        # parse the element names from the circuit string
        names = circuit.replace('d', '').replace('(', '').replace(')', '')#edit for nleis.py

        names = names.replace('p', '').replace('(', '').replace(')', '')
        names = names.replace(',', '-').replace(' ', '').split('-')

        full_names, all_units = [], []
        for name in names:
            elem = get_element_from_name(name)
            num_params = check_and_eval(elem).num_params
            units = check_and_eval(elem).units
            if num_params > 1:
                for j in range(num_params):
                    full_name = '{}_{}'.format(name, j)
                    if full_name not in constants.keys():
                        full_names.append(full_name)
                        all_units.append(units[j])
            else:
                if name not in constants.keys():
                    full_names.append(name)
                    all_units.append(units[0])

        return full_names, all_units
    def __str__(self):
        """ 
        
        Defines the pretty printing of the circuit
        
        """

        to_print = '\n'
        if self.name is not None:
            to_print += 'Name: {}\n'.format(self.name)
        to_print += 'EIS Circuit string: {}\n'.format(self.circuit_1)
        to_print += 'NLEIS Circuit string: {}\n'.format(self.circuit_2)
        to_print += "Fit: {}\n".format(self._is_fit())

        if len(self.constants) > 0:
            to_print += '\nConstants:\n'
            for name, value in self.constants.items():
                elem = get_element_from_name(name)
                units = check_and_eval(elem).units
                if '_' in name:
                    unit = units[int(name.split('_')[-1])]
                else:
                    unit = units[0]
                to_print += '  {:>5} = {:.2e} [{}]\n'.format(name, value, unit)

        names1, units1= self.get_param_names(self.circuit_1,self.constants_1)
        names2, units2 = self.get_param_names(self.circuit_2,self.constants_2)

        to_print += '\nEIS Initial guesses:\n'
        p1, p2 = individual_parameters(self.circuit_1,self.initial_guess,self.constants_1,self.constants_2)
        for name, unit, param in zip(names1, units1, p1):
            to_print += '  {:>5} = {:.2e} [{}]\n'.format(name, param, unit)
        to_print += '\nNLEIS Initial guesses:\n'
        for name, unit, param in zip(names2, units2, p2):
            to_print += '  {:>5} = {:.2e} [{}]\n'.format(name, param, unit)
        if self._is_fit():
            params1, confs1 = self.p1, self.conf1
            to_print += '\nEIS Fit parameters:\n'
            for name, unit, param, conf in zip(names1, units1, params1, confs1):
                to_print += '  {:>5} = {:.2e}'.format(name, param)
                to_print += '  (+/- {:.2e}) [{}]\n'.format(conf, unit)
            params2, confs2 = self.p2, self.conf2
            to_print += '\nNLEIS Fit parameters:\n'
            for name, unit, param, conf in zip(names2, units2, params2, confs2):
                to_print += '  {:>5} = {:.2e}'.format(name, param)
                to_print += '  (+/- {:.2e}) [{}]\n'.format(conf, unit)

        return to_print
    def extract(self):
        """ 
        
        extract the printing of the circuit in a dictionary
        
        """
    
        names1, units1 = self.get_param_names(self.circuit_1,self.constants_1)
        dic1={}
        if self._is_fit():
            params1 = self.p1
            for names, param in zip(names1, params1):
                dic1[names] = param
                
        names2, units2 = self.get_param_names(self.circuit_2,self.constants_2)
        dic2={}
        if self._is_fit():
            params2 = self.p2
            
            for names, param in zip(names2, params2):
                dic2[names] = param
    
        return dic1,dic2
    def plot(self, ax=None, f_data=None, Z1_data =None, Z2_data= None, kind='nyquist', max_f = 10, **kwargs):
        '''
        
        visualizes the model and optional data as a nyquist,
            bode, or altair (interactive) plots
        

        Parameters
        ----------
        ax : matplotlib.axes, 
            axes to plot on
        f_data : np.array of type float
           Frequencies of input data (for Bode plots)
        Z1_data : np.array of type complex
            DESCRIPTION. The default is None.
        Z2_data : np.array of type complex
            DESCRIPTION. The default is None.
        kind : {'altair', 'nyquist', 'bode'}
            type of plot to visualize
        max_f : TYPE, optional
            DESCRIPTION. The default is 10.
        **kwargs : optional
            If kind is 'nyquist' or 'bode', used to specify additional
             `matplotlib.pyplot.Line2D` properties like linewidth,
             line color, marker color, and labels.
            If kind is 'altair', used to specify nyquist height as `size`


        '''

        if kind == 'nyquist':
            if ax is None:
                _, ax = plt.subplots(1,2,figsize=(10, 5))
            
            ## we don't need the if else statement if we want to enable plot without fit 
            # if self._is_fit():
            if f_data is not None:
                f_pred = f_data
            else:
                f_pred = np.logspace(5, -3)

            if Z1_data is not None:
                ax[0] = plot_first(ax[0], Z1_data, scale=1, fmt='s', **kwargs)
                ## impedance.py style
                # plot_nyquist(Z1_data, ls='', marker='s', ax=ax[0], **kwargs)
            if Z2_data is not None:
                mask = mask =np.array(f_pred)<max_f
                ax[1] = plot_second(ax[1], Z2_data[mask], scale=1, fmt='s', **kwargs)
                ## impedance.py style
                # plot_nyquist(Z2_data, units='Ohms/A', ls='', marker='s', ax=ax[1], **kwargs)
                
            Z1_fit,Z2_fit = self.predict(f_pred,max_f=max_f)
            ax[0] = plot_first(ax[0], Z1_fit, scale=1, fmt='-', **kwargs)
            # plot_nyquist(Z1_fit, ls='-', marker='', ax=ax[0], **kwargs)
            ax[1] = plot_second(ax[1], Z2_fit, scale=1, fmt='-', **kwargs)
            # plot_nyquist(Z2_fit,units='Ohms/A', ls='-', marker='', ax=ax[1], **kwargs)
                
            ax[0].legend(['data','fit'])
            ax[1].legend(['data','fit'])
            return ax
        elif kind == 'bode':
            if ax is None:
                _, ax = plt.subplots(2,2, figsize=(8, 8))

            if f_data is not None:
                f_pred = f_data
            else:
                f_pred = np.logspace(5, -3)

            if Z1_data is not None:
                if f_data is None:
                    raise ValueError('f_data must be specified if' +
                                     ' Z_data for a Bode plot')
                ax[:,0] = plot_bode(f_data, Z1_data, ls='', marker='s',
                               axes=ax[:,0], **kwargs)
            if Z2_data is not None:
                if f_data is None:
                    raise ValueError('f_data must be specified if' +
                                     ' Z_data for a Bode plot')
                mask =np.array(f_pred)<max_f
                f2 = f_data[mask]
                Z2 = Z2_data[mask]
                ax[:,1] = plot_bode(f2,Z2, units='Ohms/A', ls='', marker='s',
                               axes=ax[:,1], **kwargs)
            ## we don't need the if else statement if we want to enable plot without fit 
            # if self._is_fit():
            Z1_fit,Z2_fit = self.predict(f_pred,max_f=max_f)
            f1 = f_data
            f2 = f_data[np.array(f_data)<max_f]

            ax[:,0] = plot_bode(f1, Z1_fit, ls='-', marker='o',
                           axes=ax[:,0], **kwargs)
            ax[:,1] = plot_bode(f2, Z2_fit,units='Ohms/A', ls='-', marker='o',
                           axes=ax[:,1], **kwargs)

            ax[0,0].set_ylabel(r'$|Z_{1}(\omega)|$ ' +
                      '$[{}]$'.format('Ohms'), fontsize=20)
            ax[1,0].set_ylabel(r'$-\phi_{Z_{1}}(\omega)$ ' + r'$[^o]$', fontsize=20)
            ax[0,1].set_ylabel(r'$|Z_{2}(\omega)|$ ' +
                      '$[{}]$'.format('Ohms/A'), fontsize=20)
            ax[1,1].set_ylabel(r'$-\phi_{Z_{2}}(\omega)$ ' + r'$[^o]$', fontsize=20)
            ax[0,0].legend(['Data','Fit'],fontsize=20)
            ax[0,1].legend(['Data','Fit'],fontsize=20)
            ax[1,0].legend(['Data','Fit'],fontsize=20)
            ax[1,1].legend(['Data','Fit'],fontsize=20)
            return ax
        elif kind == 'altair':
            plot_dict_1 = {}
            plot_dict_2 = {}


            if Z1_data is not None and Z2_data is not None and f_data is not None:
                plot_dict_1['data'] = {'f': f_data, 'Z': Z1_data}
                mask = np.array(f_data)<max_f
                plot_dict_2['data'] = {'f': f_data[mask], 'Z': Z2_data[mask]}
            ## we don't need the if else statement if we want to enable plot without fit 
            # if self._is_fit():
            if f_data is not None:
                
                f_pred = f_data
                
            else:
               
                f_pred = np.logspace(5, -3)

            if self.name is not None:
                name = self.name
            else:
                name = 'fit'
            
            Z1_fit,Z2_fit = self.predict(f_pred,max_f=max_f)
            mask = np.array(f_pred)<max_f
            plot_dict_1[name] = {'f': f_pred, 'Z': Z1_fit, 'fmt': '-'}
            plot_dict_2[name] = {'f': f_pred[mask], 'Z': Z2_fit, 'fmt': '-'}
            

            chart1 = plot_altair(plot_dict_1,k=1,units = 'Ω', **kwargs)
            chart2 = plot_altair(plot_dict_2,k=2,units = 'Ω/A', **kwargs)

            return chart1,chart2
        else:
            raise ValueError("Kind must be one of 'altair'," +
                             f"'nyquist', or 'bode' (received {kind})")
    def save(self, filepath):
        """ 
        
        Exports a model to JSON

        Parameters
        ----------
        filepath: str
            Destination for exporting model object
            
        """

        model_string_1 = self.circuit_1
        model_string_2 = self.circuit_2
        edited_circuit_str = self.edited_circuit
        model_name = self.name

        initial_guess = self.initial_guess

        if self._is_fit():
            parameters_ = list(self.parameters_)
            model_conf_ = list(self.conf_)
            # parameters_ = self.parameters_
            # model_conf_ = self.conf_
            data_dict = {"Name": model_name,
                         "Circuit String 1": model_string_1,
                         "Circuit String 2": model_string_2,
                         "Initial Guess": initial_guess,
                         "Constants": self.constants,
                         "Constants 1": self.constants_1,
                         "Constants 2": self.constants_2,
                         "Fit": True,
                         "Parameters": parameters_,
                         "Confidence": model_conf_,
                         "Edited Circuit Str": edited_circuit_str
                         }
        else:
            data_dict = {"Name": model_name,
                         "Circuit String 1": model_string_1,
                         "Circuit String 2": model_string_2,
                         "Initial Guess": initial_guess,
                         "Constants": self.constants,
                         "Constants 1": self.constants_1,
                         "Constants 2": self.constants_2,
                         "Edited Circuit Str": edited_circuit_str,
                         "Fit": False}

        with open(filepath, 'w') as f:
            json.dump(data_dict, f)


    def load(self, filepath, fitted_as_initial=False):
        """ 
        
        Imports a model from JSON

        Parameters
        ----------
        filepath: str
            filepath to JSON file to load model from

        fitted_as_initial: bool
            If true, loads the model's fitted parameters
            as initial guesses

            Otherwise, loads the model's initial and
            fitted parameters as a completed model
        
        """

        json_data_file = open(filepath, 'r')
        json_data = json.load(json_data_file)

        model_name = json_data["Name"]
        model_string_1 = json_data["Circuit String 1"]
        model_string_2 = json_data["Circuit String 2"]
        model_initial_guess = json_data["Initial Guess"]
        model_constants = json_data["Constants"]

        self.initial_guess = model_initial_guess
        self.circuit_1 = model_string_1
        self.circuit_2 = model_string_2
        self.edited_circuit = json_data["Edited Circuit Str"]

        self.constants = model_constants
        self.constants_1 =json_data["Constants 1"]
        self.constants_2 =json_data["Constants 2"]
        self.p1, self.p2 = individual_parameters(self.circuit_1,self.initial_guess,self.constants_1,self.constants_2)

        self.name = model_name

        if json_data["Fit"]:
            if fitted_as_initial:
                self.initial_guess = json_data['Parameters']
                self.p1, self.p2 = individual_parameters(self.circuit_1,self.initial_guess,self.constants_1,self.constants_2)

            else:
                self.parameters_ = np.array(json_data["Parameters"])
                self.conf_ = np.array(json_data["Confidence"])
                self.conf1, self.conf2 = individual_parameters(self.circuit_1,self.conf_,self.constants_1,self.constants_2)
                self.p1, self.p2 = individual_parameters(self.circuit_1,self.parameters_,self.constants_1,self.constants_2)

    

class NLEISCustomCircuit(BaseCircuit):
    ### this class can be fully integrated into CustomCircuit in the future
    ### , but for the stable performance of nleis.py, we overwrite it here
    def __init__(self, circuit='', **kwargs):
        """ 
        
        Constructor for a customizable equivalent circuit model

        Parameters
        ----------
        initial_guess: numpy array
            Initial guess of the circuit values

        circuit: string
            A string that should be interpreted as an equivalent circuit

        Notes
        -----
        A custom circuit is defined as a string comprised of elements in series
        (separated by a `-`) and elements in parallel (grouped as (x,y)).
        Each element can be appended with an integer (e.g. R0) or an underscore
        and an integer (e.g. CPE_1) to make keeping track of multiple elements
        of the same type easier.

        Example:
            Randles circuit is given by 'R0-p(R1-Wo1,C1)'

        """

        super().__init__(**kwargs)
        self.circuit = circuit.replace(" ", "")

        circuit_len = calculateCircuitLength(self.circuit)

        if len(self.initial_guess) + len(self.constants) != circuit_len:
            raise ValueError('The number of initial guesses ' +
                             f'({len(self.initial_guess)}) + ' +
                             'the number of constants ' +
                             f'({len(self.constants)})' +
                             ' must be equal to ' +
                             f'the circuit length ({circuit_len})')
    def fit(self, frequencies, impedance, bounds=None,
            weight_by_modulus=False, max_f =10, **kwargs):
        '''
        
        Fit the circuit model

        Parameters
        ----------
        frequencies : numpy array
            Frequencies.
        impedance : numpy array of dtype 'complex128'
            Impedance values to fit.
        bounds : 2-tuple of array_like, optional
            Lower and upper bounds on parameters.
        weight_by_modulus : bool, optional
            Uses the modulus of each data (`|Z|`) as the weighting factor.
            Standard weighting scheme when experimental variances are
            unavailable. Only applicable when global_opt = False
        max_f : float , 10
            Truncation for 2nd-NLEIS based on previous experiments. The default is 10.
        **kwargs : 
            Keyword arguments passed to
            impedance.models.circuits.fitting.circuit_fit,
            and subsequently to scipy.optimize.curve_fit
            or scipy.optimize.basinhopping.

        Raises
        ------
        TypeError
            DESCRIPTION.
        ValueError
            DESCRIPTION.

        Returns
        -------
        self: returns an instance of self

        '''
            
        frequencies = np.array(frequencies, dtype=float)
        impedance = np.array(impedance, dtype=complex)

        if len(frequencies) != len(impedance):
            raise TypeError('length of frequencies and impedance do not match')
        mask = np.array(frequencies)<max_f
        frequencies = frequencies[mask]
        impedance = impedance[mask]

        if self.initial_guess != []:
            parameters, conf = circuit_fit(frequencies, impedance,
                                           self.circuit, self.initial_guess,
                                           constants=self.constants,
                                           bounds=bounds,
                                           weight_by_modulus=weight_by_modulus,
                                           **kwargs)
            self.parameters_ = parameters
            if conf is not None:
                self.conf_ = conf
        else:
            raise ValueError('No initial guess supplied')

        return self
    def predict(self, frequencies, use_initial=False):
        """ 
        
        Predict impedance using an equivalent circuit model

        Parameters
        ----------
        frequencies: array-like of numeric type
        use_initial: boolean
            If true and the model was previously fit use the initial
            parameters instead

        Returns
        -------
        impedance: ndarray of dtype 'complex128'
            Predicted impedance at each frequency
        
        """
        frequencies = np.array(frequencies, dtype=float)

        if self._is_fit() and not use_initial:
            return eval(buildCircuit(self.circuit, frequencies,
                                     *self.parameters_,
                                     constants=self.constants, eval_string='',
                                     index=0)[0],
                        circuit_elements)
        else:
            warnings.warn("Simulating circuit based on initial parameters")
            return eval(buildCircuit(self.circuit, frequencies,
                                     *self.initial_guess,
                                     constants=self.constants, eval_string='',
                                     index=0)[0],
                        circuit_elements)
    def get_param_names(self):
        """ 
        
        Converts circuit string to names and units 
        
        """

        # parse the element names from the circuit string
        names = self.circuit.replace('d', '').replace('(', '').replace(')', '')#edited for nleis.py

        names = names.replace('p', '').replace('(', '').replace(')', '')
        names = names.replace(',', '-').replace(' ', '').split('-')

        full_names, all_units = [], []
        for name in names:
            elem = get_element_from_name(name)
            num_params = check_and_eval(elem).num_params
            units = check_and_eval(elem).units
            if num_params > 1:
                for j in range(num_params):
                    full_name = '{}_{}'.format(name, j)
                    if full_name not in self.constants.keys():
                        full_names.append(full_name)
                        all_units.append(units[j])
            else:
                if name not in self.constants.keys():
                    full_names.append(name)
                    all_units.append(units[0])

        return full_names, all_units
    def extract(self):
        """ 
        
        extract the printing of the circuit 
        
        """

        names, units = self.get_param_names()
        dic={}
        if self._is_fit():
            params, confs = self.parameters_, self.conf_
            
            for name, unit, param, conf in zip(names, units, params, confs):
                dic[name] = param

        return dic

    



import numpy as np

from impedance.preprocessing import ignoreBelowX

from impedance.tests.test_preprocessing import frequencies \
   as example_frequencies
from impedance.tests.test_preprocessing import Z_correct

from nleis.fitting import buildCircuit, \
   circuit_fit, mape,mae, extract_circuit_elements, \
   set_default_bounds, seq_fit_parm
from nleis.nleis_fitting import data_processing, \
simul_fit,individual_parameters
import os
test_dir = os.path.dirname(os.path.abspath(__file__))
data_dir = os.path.join(test_dir, '../data')

def test_data_processing():

    # # get example data
    # # The example is shown in "Getting Started" page 
    
    frequencies = np.loadtxt(os.path.join(data_dir, 'freq_30a.txt'))
    Z1 = np.loadtxt(os.path.join(data_dir,'Z1s_30a.txt')).view(complex)[1]
    Z2 = np.loadtxt(os.path.join(data_dir,'Z2s_30a.txt')).view(complex)[1]

    max_f = 10

    f,Z1,Z2,f2_truncated,Z2_truncated = data_processing(frequencies,Z1,Z2,max_f=max_f)
    assert len(f)==len(Z1)==len(Z2)
    assert Z1.imag.max() < 0 
    assert f2_truncated.max()<max_f
    assert len(f2_truncated) == len(Z2_truncated)



def test_set_default_bounds():
    # Test example circuit that calculate the difference between
    # two porous electrode with spherical particle and with platelet like particle 
    circuit = 'd(TDSn0,TDPn1)'

    # Test with no constants
    default_bounds = (np.zeros(14), np.inf*np.ones(14))
    default_bounds[0][5] = -np.inf
    default_bounds[0][6] = -0.5
    default_bounds[1][6] = 0.5
    default_bounds[0][12] = -np.inf
    default_bounds[0][13] = -0.5
    default_bounds[1][13] = 0.5

    bounds_from_func = set_default_bounds(circuit)

    assert np.allclose(default_bounds, bounds_from_func)

    # Test with constants
    constants = {'TDSn0_0': 1}
    default_bounds = (np.zeros(13), np.inf*np.ones(13))
    default_bounds[0][4] = -np.inf
    default_bounds[0][5] = -0.5
    default_bounds[1][5] = 0.5
    default_bounds[0][11] = -np.inf
    default_bounds[0][12] = -0.5
    default_bounds[1][12] = 0.5
    bounds_from_func = set_default_bounds(circuit, constants=constants)

    assert np.allclose(default_bounds, bounds_from_func)

#### This test remain unchanged as we are adopting it from impedance.py
def test_circuit_fit():

    # Test trivial model (10 Ohm resistor)
    circuit = 'R0'
    initial_guess = [10]

    results_simple = [10]

    frequencies = np.array([10, 100, 1000])
    Z_data = np.array([10, 10, 10])  # impedance is real

    assert np.allclose(circuit_fit(frequencies, Z_data, circuit,
                                   initial_guess, constants={},
                                   global_opt=True)[0],
                       results_simple, rtol=1e-1)

    # check that list inputs work
    frequency_list = [10, 100, 1000]
    Z_data_list = [10, 10, 10]

    assert np.allclose(circuit_fit(frequency_list, Z_data_list, circuit,
                                   initial_guess, constants={},
                                   global_opt=True)[0],
                       results_simple, rtol=1e-1)

    # Test example circuit from "Getting Started" page
    circuit = 'R0-p(R1,C1)-p(R2-Wo1,C2)'
    initial_guess = [.01, .01, 100, .01, .05, 100, 1]
    bounds = [(0, 0, 0, 0, 0, 0, 0),
              (10, 1, 1e3, 1, 1, 1e4, 100)]

    # results change slightly using predefined bounds
    results_local = np.array([1.65e-2, 8.68e-3, 3.32, 5.39e-3,
                              6.31e-2, 2.33e2, 2.20e-1])
    results_local_bounds = results_local.copy()
    results_local_bounds[5] = 2.38e2
    results_local_weighted = np.array([1.64e-2, 9.06e-3, 3.06,
                                       5.29e-3, 1.45e-1, 1.32e3, 2.02e-1])

    results_global = np.array([1.65e-2, 5.34e-3, 0.22, 9.15e-3,
                               1.31e-1, 1.10e3, 2.78])

    # Filter
    example_frequencies_filtered, \
        Z_correct_filtered = ignoreBelowX(example_frequencies, Z_correct)

    # Test local fitting
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, constants={})[0],
                       results_local, rtol=1e-2)

    # Test local fitting with predefined bounds
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, bounds=bounds,
                                   constants={})[0],
                       results_local_bounds, rtol=1e-2)

    # Test local fitting with predefined weights
    # Use abs(Z), stacked in order of (Re, Im) components
    sigma = np.hstack((np.abs(Z_correct_filtered),
                       np.abs(Z_correct_filtered)))
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, sigma=sigma,
                                   constants={})[0],
                       results_local_weighted, rtol=1e-2)

    # Test if using weight_by_modulus=True produces the same results
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, weight_by_modulus=True,
                                   constants={})[0],
                       results_local_weighted, rtol=1e-2)

    # Test global fitting on multiple seeds
    # All seeds should converge to the same parameter values
    # seed = 0 (default)
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, constants={},
                                   global_opt=True)[0],
                       results_global, rtol=1e-1)

    # seed = 0, with predefined bounds
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, constants={},
                                   global_opt=True, bounds=bounds,
                                   seed=0)[0],
                       results_global, rtol=1e-1)

    # seed = 1
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, constants={},
                                   global_opt=True, seed=1)[0],
                       results_global, rtol=1e-1)

    # seed = 42
    assert np.allclose(circuit_fit(example_frequencies_filtered,
                                   Z_correct_filtered, circuit,
                                   initial_guess, constants={},
                                   global_opt=True, seed=42)[0],
                       results_global, rtol=1e-1)
    

def test_buildCircuit():

   # Test a simple porous electrode with contact resistor and inductance
   circuit = 'L0-R0-TDS0'
   params = [1e-5, 0.01, 1, 0.1, 10, 1, 1000]
   frequencies = [1000.0, 5.0, 0.01]

   assert buildCircuit(circuit, frequencies, *params,
                       constants={})[0].replace(' ', '') == \
       's([L([1e-05],[1000.0,5.0,0.01]),' + \
       'R([0.01],[1000.0,5.0,0.01]),' + \
       'TDS([1.0,0.1,10.0,1.0,1000.0],[1000.0,5.0,0.01])])' 

# Test the substraction between two elements
   circuit = 'd(R1,R2)'
   params = [0.1, 0.2]
   frequencies = [1000.0, 5.0, 0.01]

   assert buildCircuit(circuit, frequencies, *params,
                       constants={})[0].replace(' ', '') == \
       'd([R([0.1],[1000.0,5.0,0.01]),' + \
       'R([0.2],[1000.0,5.0,0.01])])'

   # Test nested parallel and difference groups
   circuit = 'R0-d(p(R1, C1)-R2, C2)'
   params = [1, 2, 3, 4, 5]
   frequencies = [1000.0, 5.0, 0.01]

   assert buildCircuit(circuit, frequencies, *params,
                       constants={})[0].replace(' ', '') == \
       's([R([1],[1000.0,5.0,0.01]),' + \
       'd([s([p([R([2],[1000.0,5.0,0.01]),' + \
       'C([3],[1000.0,5.0,0.01])]),' + \
       'R([4],[1000.0,5.0,0.01])]),' + \
       'C([5],[1000.0,5.0,0.01])])])'


   # Test single element circuit
   circuit = 'R1'
   params = [100]
   frequencies = [1000.0, 5.0, 0.01]

   assert buildCircuit(circuit, frequencies, *params,
                       constants={})[0].replace(' ', '') == \
       'R([100],[1000.0,5.0,0.01])'


def test_mae():
   a = np.array([2 + 4*1j, 3 + 2*1j])
   b = np.array([2 + 4*1j, 3 + 2*1j])

   assert mae(a, b) == 0.0

   c = np.array([2 + 4*1j, 1 + 4*1j])
   d = np.array([4 + 2*1j, 3 + 2*1j])
   assert np.isclose(mae(c, d), np.sqrt(8))

def test_mape():
   a = np.array([2 + 4*1j, 3 + 2*1j])
   b = np.array([2 + 4*1j, 3 + 2*1j])

   assert mape(a, b) == 0.0

   c = np.array([2 + 4*1j, 1 + 4*1j])
   d = np.array([4 + 2*1j, 3 + 2*1j])
   assert np.isclose(mape(c, d), 0.5*np.sqrt(8)*(1/np.sqrt(20)+1/np.sqrt(17))*100)

def test_element_extraction():
   circuit = 'd(TDSn0,TDPn1)'
   extracted_elements = extract_circuit_elements(circuit)
   assert extracted_elements == ['TDSn0', 'TDPn1']

def test_seq_fit_parm():
   input_dic = {'TDS0_1':0,'TDS1_3':0}
   target_arr = ['TDS0','TDS1']
   output_arr = ['TDSn0','TDSn1']
   output_dic = seq_fit_parm(input_dic,target_arr,output_arr)
   assert output_dic == {'TDSn0_1':0,'TDSn1_3':0}


def test_simul_fit():
    # get example data
    # The example is shown in "Getting Started" page 
    
    frequencies = np.loadtxt(os.path.join(data_dir, 'freq_30a.txt'))
    Z1 = np.loadtxt(os.path.join(data_dir,'Z1s_30a.txt')).view(complex)[1]
    Z2 = np.loadtxt(os.path.join(data_dir,'Z2s_30a.txt')).view(complex)[1]

    circ_str_1 = 'L0-R0-TDS0-TDS1'
    circ_str_2 = 'd(TDSn0,TDSn1)'
    edited_circuit = 'L0-R0-TDSn0-TDSn1'

   #  initial_guess = [1e-7,1e-3 # L0,RO
   #                  ,5e-3,1e-3,10,1e-2,100,10,0.1 ## TDS0 + additioal nonlinear parameters
   #                  ,1e-3,1e-3,1e-3,1e-2,1000,0,0 ## TDS1 + additioal nonlinear parameters
   #                  ]
   ## The test framework here can be improved in the future after getting much stable initial guess
    results =[9.81376136e-08,1.34551860e-02,2.52398185e-02,5.06155989e-03,
         8.82298681,8.78424359e-05,3.58691184,1.22575311e+01,
         8.74768343e-02,2.09036506e-02,1.13813146e-03,8.13615357e-01,
         1.86194200e+02,3.29019412e+09,1.02283674,6.38865883e-03]
    initial_guess= results
    
    p, perror =  simul_fit(frequencies, Z1, Z2, circ_str_1,circ_str_2, edited_circuit, initial_guess, constants_1={},constants_2={},
                bounds = None, opt='max',cost = 0.5,max_f=10,param_norm = True,positive = True)

    assert np.allclose (p,results)

   
def test_individual_parameters():

    ## EIS and 2nd-NLEIS circ_str are defined here to make sure 
    ## we know we have a a pair of consistent models 

    circ_str_1 = 'L0-R0-TDS0-TDS1'
    circ_str_2 = 'd(TDSn0,TDSn1)'

    initial_guess = [1e-7,1e-3 # L0,RO
                    ,5e-3,1e-3,10,1e-2,100,10,0.1 ## TDS0 + additioal nonlinear parameters
                    ,1e-3,1e-3,1e-3,1e-2,1000,0,0 ## TDS1 + additioal nonlinear parameters
                    ]
    ## Test without constant
    p1, p2 = individual_parameters(circ_str_1,initial_guess,constants_1 = {},constants_2= {})
    assert np.allclose(p1,[1e-7,1e-3
                    ,5e-3,1e-3,10.0,1e-02,100.0
                    ,1e-3,1e-03,1e-03,1e-02,1000.0])
    assert np.allclose(p2,[5e-3,1e-3,10,1e-2,100,10,0.1
                    ,1e-3,1e-3,1e-3,1e-2,1000,0,0])
    
    ## Test with constant
    initial_guess = [1e-7,1e-3 # L0,RO
                ,5e-3,10,1e-2,100,10,0.1 ## TDS0 + additioal nonlinear parameters
                ,1e-3,1e-3,1e-3,1e-2,1000,0 ## TDS1 + additioal nonlinear parameters
                ]
    p1, p2 = individual_parameters(circ_str_1,initial_guess
                                   ,constants_1 = {'TDS0_1':0},constants_2= {'TDSn1_6':0})

    assert np.allclose(p1,[1e-7,1e-3
                    ,5e-3,10,1e-02,100.0
                    ,1e-3,1e-03,1e-03,1e-02,1000.0])
    assert np.allclose(p2,[5e-3,10,1e-2,100,10,0.1
                    ,1e-3,1e-3,1e-3,1e-2,1000,0])




import warnings

import numpy as np
from scipy.linalg import inv
from scipy.optimize import curve_fit, basinhopping
from impedance.models.circuits.elements import circuit_elements, get_element_from_name
from impedance.models.circuits.fitting import check_and_eval,rmse
from .fitting import set_default_bounds,buildCircuit,calculateCircuitLength,extract_circuit_elements
from scipy.optimize import minimize
ints = '0123456789'

def data_processing(f,Z1,Z2,max_f=10):
    mask = np.array(Z1.imag)<0
    f = f[mask]
    Z1 = Z1[mask]
    Z2 = Z2[mask]
    mask1 = np.array(f)<max_f
    f2_truncated = f[mask1]
    Z2_truncated = Z2 [mask1]
    return (f,Z1,Z2,f2_truncated,Z2_truncated)

def simul_fit(frequencies, Z1, Z2, circuit_1,circuit_2, edited_circuit, initial_guess, constants_1={},constants_2={},
                bounds = None, opt='max',cost = 0.5,max_f=10,param_norm = True,positive = True,
                **kwargs):

    """ Main function for the simultaneous fitting of EIS and NLEIS edata.

    By default, this function uses `scipy.optimize.curve_fit
    <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html>`_
    to fit the equivalent circuit. 

    Parameters
    ----------
    frequencies : numpy array
        Frequencies

    Z1 : numpy array of dtype 'complex128'
        EIS
    Z1 : numpy array of dtype 'complex128'
        NLEIS

    circuit_1 : string
        String defining the EIS equivalent circuit to be fit
    circuit_2 : string
        String defining the NLEIS equivalent circuit to be fit

    initial_guess : list of floats
        Initial guesses for the fit parameters

    constants : dictionary, optional
        Parameters and their values to hold constant during fitting
        (e.g. {"RO": 0.1}). Defaults to {}

    bounds : 2-tuple of array_like, optional
        Lower and upper bounds on parameters. Defaults to bounds on all
        parameters of 0 and np.inf, except the CPE alpha
        which has an upper bound of 1

    opt : str, optional
        Default is max normalization. Other normalization will be supported
        in the future 
    cost : float, default = 0.5
        cost function: cost > 0.5 means more weight on EIS while cost < 0.5 means more weight on NLEIS

    max_f: int
        The the maximum frequency of interest for NLEIS
        
    positive : bool, optional
        Defaults to True for only positive nyquist plot
        
    param_norm : bool, optional
         Defaults to True for better convergence
    

    kwargs :
        Keyword arguments passed to scipy.optimize.curve_fit or
        scipy.optimize.basinhopping

    Returns
    -------
    p_values : list of floats
        best fit parameters for EIS and NLEIS data

    p_errors : list of floats
        one standard deviation error estimates for fit parameters

    """
    ###  Todo fix the negtive loglikelihood
  
    ### set upper and lower bounds on a per-element basis

    if bounds is None:
        combined_constant = constants_2.copy()
        combined_constant.update(constants_1)
        bounds = set_default_bounds(edited_circuit, constants=combined_constant)
        ub = np.ones(len(bounds[1]))
    else:
        if param_norm:
            ub = bounds[1]
            bounds = bounds/ub
        else:
            ub = np.ones(len(bounds[1]))
            
    initial_guess = initial_guess/ub


    if positive:
        mask1 = np.array(Z1.imag)<0
        frequencies = frequencies[mask1]
        Z1 = Z1[mask1]
        Z2 = Z2[mask1]
        mask2 = np.array(frequencies)<max_f
        Z2 = Z2[mask2] 
    else:
        mask2 = np.array(frequencies)<max_f
        Z2 = Z2[mask2] 

    Z1stack = np.hstack([Z1.real, Z1.imag])
    Z2stack = np.hstack([Z2.real, Z2.imag])
    Zstack = np.hstack([Z1stack,Z2stack])
    # weighting scheme for fitting
    if opt == 'max':
        if 'maxfev' not in kwargs:
            kwargs['maxfev'] = 1e5
        if 'ftol' not in kwargs:
            kwargs['ftol'] = 1e-13
        Z1max = max(np.abs(Z1))
        Z2max = max(np.abs(Z2))
        
        sigma1 = np.ones(len(Z1stack))*Z1max/(cost**0.5)
        sigma2 = np.ones(len(Z2stack))*Z2max/((1-cost)**0.5)
        kwargs['sigma'] = np.hstack([sigma1, sigma2])

        popt, pcov = curve_fit(wrapCircuit_simul(circuit_1, constants_1,circuit_2,constants_2,ub,max_f), frequencies,
                           Zstack,
                           p0=initial_guess, bounds=bounds, **kwargs)
    

    # Calculate one standard deviation error estimates for fit parameters,
    # defined as the square root of the diagonal of the covariance matrix.
    # https://stackoverflow.com/a/52275674/5144795
    # and the following for the bounded and normalized case
    # https://stackoverflow.com/questions/14854339/in-scipy-how-and-why-does-curve-fit-calculate-the-covariance-of-the-parameter-es
        perror = np.sqrt(np.diag(ub*pcov*ub.T))

        return popt*ub, perror
    if opt == 'neg':
        ### This method does not provides converge solution at current development
        bounds = tuple(tuple((bounds[0][i], bounds[1][i])) for i in range(len(bounds[0])))

        res = minimize(wrapNeg_log_likelihood(frequencies,Z1,Z2,circuit_1, constants_1,circuit_2,constants_2,ub,max_f,cost = cost), x0=initial_guess,bounds=bounds, **kwargs)
        
        return (res.x*ub,None)
    

def wrapNeg_log_likelihood(frequencies,Z1,Z2,circuit_1, constants_1,circuit_2,constants_2,ub,max_f=10,cost=0.5):
    ''' wraps function for negtive log likelihood optimization'''
    
    def wrappedNeg_log_likelihood(parameters):
        """ returns a stacked array of real and imaginary impedance
        components

        Parameters
        ----------
        frequencies: list of floats
        Z1: EIS data
        Z2: NLEIS data
        circuit_1 : string        
        constants_1 : dict
        circuit_2 : string        
        constants_2 : dict
        ub : list of floats upper bound if bounds are provided
        max_f: int
        parameters : list of floats

        Returns
        -------
        array of floats

        """
        f1 =frequencies
        mask = np.array(frequencies)<max_f
        f2 = frequencies[mask]
        x1,x2 = wrappedImpedance(circuit_1, constants_1,circuit_2,constants_2,f1,f2,parameters*ub)
        # Z1max = max(np.abs(Z1))
        # Z2max = max(np.abs(Z2))
        # log1 = np.log(sum(((Z1.real-x1.real)/Z1max)**2))+np.log(sum(((Z1.imag-x1.imag)/Z1max)**2))
        # log2 = np.log(sum(((Z2.real-x2.real)/Z2max)**2))+np.log(sum(((Z2.imag-x2.imag)/Z2max)**2))
        log1 = np.log(sum(((Z1-x1))**2))
        log2 = np.log(sum(((Z2-x2))**2))
        return(cost*log1+(1-cost)*log2)
    return wrappedNeg_log_likelihood
        
        


def wrapCircuit_simul(circuit_1, constants_1,circuit_2,constants_2,ub,max_f=10):
    """ wraps function so we can pass the circuit string """
    def wrappedCircuit_simul(frequencies, *parameters):
        """ returns a stacked array of real and imaginary impedance
        components

        Parameters
        ----------
        circuit_1 : string        
        constants_1 : dict
        circuit_2 : string        
        constants_2 : dict
        max_f: int
        parameters : list of floats
        frequencies : list of floats

        Returns
        -------
        array of floats

        """
        
        f1 =frequencies
        mask = np.array(frequencies)<max_f
        f2 = frequencies[mask]
        x1,x2 = wrappedImpedance(circuit_1, constants_1,circuit_2,constants_2,f1,f2,parameters*ub)

        y1_real = np.real(x1)
        y1_imag = np.imag(x1)
        y1_stack = np.hstack([y1_real, y1_imag])
        y2_real = np.real(x2)
        y2_imag = np.imag(x2)
        y2_stack = np.hstack([y2_real, y2_imag])

        return np.hstack([y1_stack, y2_stack])
    return wrappedCircuit_simul

def wrappedImpedance(circuit_1, constants_1,circuit_2,constants_2,f1,f2,parameters):
    '''

    Parameters
    ----------
    circuit_1 : string
    constants_1 : dict
    circuit_2 : string
    constants_2 : dict
    f1 : list of floats
    f2 : list of floats
    parameters : list of floats


    Returns
    -------
    Z1 and Z2

    '''

    p1,p2 = individual_parameters(circuit_1,parameters,constants_1,constants_2)
    
    x1 = eval(buildCircuit(circuit_1, f1, *p1,
                          constants=constants_1, eval_string='',
                          index=0)[0],
             circuit_elements)
    x2 = eval(buildCircuit(circuit_2, f2, *p2,
                          constants=constants_2, eval_string='',
                          index=0)[0],
             circuit_elements)
    return(x1,x2)
def individual_parameters(circuit,parameters,constants_1,constants_2):
    '''

    Parameters
    ----------
    circuit : string
        DESCRIPTION.
    parameters : list of floats
        DESCRIPTION.
    constants_1 : dict
        DESCRIPTION.
    constants_2 : dict
        DESCRIPTION.

    Returns
    -------
    TYPE
        DESCRIPTION.
    TYPE
        DESCRIPTION.

    '''

    if circuit == '':
        return [],[]
    parameters = list(parameters)
    elements_1 = extract_circuit_elements(circuit)
    p1 = []
    p2 = []
    index = 0
    for elem in elements_1:
        
        raw_elem = get_element_from_name(elem)
        elem_number_1 = check_and_eval(raw_elem).num_params
        if elem[0]=='T' or elem[0:2] =='RC' : ### this might be improvable, but depends on how we want to define the name
            ## check for nonlinear element
            elem_number_2 = check_and_eval(raw_elem+'n').num_params
            
        else: 
            elem_number_2 = elem_number_1
        
        for j in range(elem_number_2):
            if elem_number_1 > 1:
                if j<elem_number_1:
                    current_elem_1 = elem + '_{}'.format(j)
                else:
                    current_elem_1 = None
                len_elem = len(raw_elem)    
                current_elem_2 = elem[0:len_elem]+'n'+elem[len_elem:] + '_{}'.format(j)
            else:
                current_elem_1 = elem
                current_elem_2 = elem
            if current_elem_1 in constants_1.keys()  :
                continue
            elif current_elem_2 in constants_2.keys() and current_elem_1 not in constants_1.keys():
                continue
            else:
                if elem_number_1 ==1:    
                    p1.append(parameters[index])
                    
                elif elem_number_1>1 and j<elem_number_1: 
                    p1.append(parameters[index])
                    p2.append(parameters[index])
                else:
                    p2.append(parameters[index])
                
                index += 1

    return p1,p2












import warnings

import numpy as np
from scipy.linalg import inv
from scipy.optimize import curve_fit, basinhopping
from impedance.models.circuits.elements import circuit_elements, get_element_from_name
from impedance.models.circuits.fitting import check_and_eval,rmse

### Note: a lot of codes are directly adopted from impedance.py. Can be easily perform a full integration in the future, but now we are keep them separated to ensure the stable performance

ints = '0123456789'


def mae(a,b):
    ''' 
   
    Mean Absolute Error

    Parameters
    ----------
    a : numpy array
        experimental data.
    b : numpy array
        model fit.

    Returns
    -------
    The calculated Mean Absolute Error.

    '''
    return(np.mean(abs(a-b)))

def mape(a,b):
    '''
   
    Mean Absolute Percentage Error

    Parameters
    ----------
    a : numpy array
        experimental data.
    b : numpy array
        model fit.

    Returns
    -------
    Mean Absolute Percentage Error.

    '''
    return(np.mean(abs(a-b)/abs(a))*100)

def seq_fit_parm(input_dic,target_arr,output_arr):
    '''
     
    Convert obtained EIS result to a constant of dictionary for 2nd-NLEIS analyssis
    for sequential optimization disscussed in [1]

    Parameters
    ----------
    input_dic : dictionary
        DESCRIPTION.
    target_arr : list of string
        DESCRIPTION.
    output_arr : list of string 
        DESCRIPTION.

    Raises
    ------
    ValueError
        DESCRIPTION.

    Returns
    -------
    None.

    '''
    output_dic = {}
    n = len(target_arr)
    if n != len(output_arr):
        raise ValueError('Target Array and Output Array Must Have the Same Length ')

    for i in range(0,n):
        for j in range(check_and_eval(target_arr[i][0:-1]).num_params):
            if target_arr[i]+'_'+str(j) not in input_dic:
                continue
            output_dic[output_arr[i]+'_'+str(j)]=input_dic[target_arr[i]+'_'+str(j)]
    return(output_dic)
    
def set_default_bounds(circuit, constants={}):
    """ 
    
    This function sets default bounds for optimization.

    set_default_bounds sets bounds of 0 and np.inf for all parameters,
    except the CPE and La alphas which have an upper bound of 1.

    Parameters
    ----------
    circuit : string
        String defining the equivalent circuit to be fit

    constants : dictionary, optional
        Parameters and their values to hold constant during fitting
        (e.g. {"RO": 0.1}). Defaults to {}

    Returns
    -------
    bounds : 2-tuple of array_like
        Lower and upper bounds on parameters.
        
    """

    # extract the elements from the circuit
    extracted_elements = extract_circuit_elements(circuit)

    # loop through bounds
    lower_bounds, upper_bounds = [], []
    
    for elem in extracted_elements:

        raw_element = get_element_from_name(elem)
        
        for i in range(check_and_eval(raw_element).num_params):
            
            if elem in constants or elem + f'_{i}' in constants:
                
                continue
                
            elif raw_element in ['CPE', 'La'] and i == 1:
                upper_bounds.append(1)
                lower_bounds.append(0)
            ## The following are for nleis.py
            elif raw_element in ['Tsn'] and i == 4:
                upper_bounds.append(0.5)
                lower_bounds.append(-0.5)
            elif raw_element in ['TPn'] and i == 3:
                upper_bounds.append(0.5)
                lower_bounds.append(-0.5)
            elif raw_element in ['RCn','RCOn'] and i == 2:
                upper_bounds.append(0.5)
                lower_bounds.append(-0.5)
            elif raw_element in ['TDSn','TDPn','TDCn'] and (i == 5 ) :

                upper_bounds.append(np.inf)
                lower_bounds.append(-np.inf)

            elif raw_element in ['TDSn','TDPn','TDCn'] and i == 6 :
                upper_bounds.append(0.5)
                lower_bounds.append(-0.5)

            elif raw_element in ['RCDn','RCSn'] and (i == 4) :
                upper_bounds.append(np.inf)
                lower_bounds.append(-np.inf)

            elif raw_element in ['RCDn','RCSn'] and i == 5 :
                upper_bounds.append(0.5)
                lower_bounds.append(-0.5)
            elif raw_element in ['TLMn'] and (i == 6 or i ==7):
                upper_bounds.append(0.5)
                lower_bounds.append(-0.5)

            elif raw_element in ['TLMSn'] and (i == 9 or i ==10):
                upper_bounds.append(0.5)
                lower_bounds.append(-0.5)
            elif raw_element in ['TLMSn'] and (i == 8):
                upper_bounds.append(np.inf)
                lower_bounds.append(-np.inf)            

            else:
                upper_bounds.append(np.inf)
                lower_bounds.append(0)
                


    bounds = ((lower_bounds), (upper_bounds))
    

    return bounds



def circuit_fit(frequencies, impedances, circuit, initial_guess, constants={},
                bounds=None, weight_by_modulus=False, global_opt=False,
                **kwargs):

    """ Main function for fitting an equivalent circuit to data.

    By default, this function uses `scipy.optimize.curve_fit
    <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html>`_
    to fit the equivalent circuit. This function generally works well for
    simple circuits. However, the final results may be sensitive to
    the initial conditions for more complex circuits. In these cases,
    the `scipy.optimize.basinhopping
    <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.basinhopping.html>`_
    global optimization algorithm can be used to attempt a better fit.

    Parameters
    ----------
    frequencies : numpy array
        Frequencies

    impedances : numpy array of dtype 'complex128'
        Impedances

    circuit : string
        String defining the equivalent circuit to be fit

    initial_guess : list of floats
        Initial guesses for the fit parameters

    constants : dictionary, optional
        Parameters and their values to hold constant during fitting
        (e.g. {"RO": 0.1}). Defaults to {}

    bounds : 2-tuple of array_like, optional
        Lower and upper bounds on parameters. Defaults to bounds on all
        parameters of 0 and np.inf, except the CPE alpha
        which has an upper bound of 1

    weight_by_modulus : bool, optional
        Uses the modulus of each data (`|Z|`) as the weighting factor.
        Standard weighting scheme when experimental variances are unavailable.
        Only applicable when global_opt = False

    global_opt : bool, optional
        If global optimization should be used (uses the basinhopping
        algorithm). Defaults to False

    kwargs :
        Keyword arguments passed to scipy.optimize.curve_fit or
        scipy.optimize.basinhopping

    Returns
    -------
    p_values : list of floats
        best fit parameters for specified equivalent circuit

    p_errors : list of floats
        one standard deviation error estimates for fit parameters

    Notes
    -----
    Need to do a better job of handling errors in fitting.
    Currently, an error of -1 is returned.

    """
    
    f = np.array(frequencies, dtype=float)
    Z = np.array(impedances, dtype=complex)

    # set upper and lower bounds on a per-element basis
    if bounds is None:
        bounds = set_default_bounds(circuit, constants=constants)

    if not global_opt:
        if 'maxfev' not in kwargs:
            kwargs['maxfev'] = int(1e5)
        if 'ftol' not in kwargs:
            kwargs['ftol'] = 1e-13

        # weighting scheme for fitting
        if weight_by_modulus:
            abs_Z = np.abs(Z)
            kwargs['sigma'] = np.hstack([abs_Z, abs_Z])

        popt, pcov = curve_fit(wrapCircuit(circuit, constants), f,
                               np.hstack([Z.real, Z.imag]),
                               p0=initial_guess, bounds=bounds, **kwargs)

        # Calculate one standard deviation error estimates for fit parameters,
        # defined as the square root of the diagonal of the covariance matrix.
        # https://stackoverflow.com/a/52275674/5144795
        perror = np.sqrt(np.diag(pcov))

    else:
        if 'seed' not in kwargs:
            kwargs['seed'] = 0

        def opt_function(x):
            """ Short function for basinhopping to optimize over.
            We want to minimize the RMSE between the fit and the data.

            Parameters
            ----------
            x : args
                Parameters for optimization.

            Returns
            -------
            function
                Returns a function (RMSE as a function of parameters).
            """
            return rmse(wrapCircuit(circuit, constants)(f, *x),
                        np.hstack([Z.real, Z.imag]))

        class BasinhoppingBounds(object):
            """ Adapted from the basinhopping documetation
            https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.basinhopping.html
            """

            def __init__(self, xmin, xmax):
                self.xmin = np.array(xmin)
                self.xmax = np.array(xmax)

            def __call__(self, **kwargs):
                x = kwargs['x_new']
                tmax = bool(np.all(x <= self.xmax))
                tmin = bool(np.all(x >= self.xmin))
                return tmax and tmin

        basinhopping_bounds = BasinhoppingBounds(xmin=bounds[0],
                                                 xmax=bounds[1])
        results = basinhopping(opt_function, x0=initial_guess,
                               accept_test=basinhopping_bounds, **kwargs)
        popt = results.x

        # Calculate perror
        jac = results.lowest_optimization_result['jac'][np.newaxis]
        try:
            # jacobian -> covariance
            # https://stats.stackexchange.com/q/231868
            pcov = inv(np.dot(jac.T, jac)) * opt_function(popt) ** 2
            # covariance -> perror (one standard deviation
            # error estimates for fit parameters)
            perror = np.sqrt(np.diag(pcov))
        except (ValueError, np.linalg.LinAlgError):
            warnings.warn('Failed to compute perror')
            perror = None

    return popt, perror

def wrapCircuit(circuit, constants):
    """ wraps function so we can pass the circuit string """
    def wrappedCircuit(frequencies, *parameters):
        """ returns a stacked array of real and imaginary impedance
        components

        Parameters
        ----------
        circuit : string
        constants : dict
        parameters : list of floats
        frequencies : list of floats

        Returns
        -------
        array of floats

        """

        x = eval(buildCircuit(circuit, frequencies, *parameters,
                              constants=constants, eval_string='',
                              index=0)[0],
                 circuit_elements)
        y_real = np.real(x)
        y_imag = np.imag(x)

        return np.hstack([y_real, y_imag])
    return wrappedCircuit



def buildCircuit(circuit, frequencies, *parameters,
                 constants=None, eval_string='', index=0):
    """ recursive function that transforms a circuit, parameters, and
    frequencies into a string that can be evaluated

    Parameters
    ----------
    circuit: str
    frequencies: list/tuple/array of floats
    parameters: list/tuple/array of floats
    constants: dict

    Returns
    -------
    eval_string: str
        Python expression for calculating the resulting fit
    index: int
        Tracks parameter index through recursive calling of the function
    
    """

    parameters = np.array(parameters).tolist()
    frequencies = np.array(frequencies).tolist()
    circuit = circuit.replace(' ', '')

    def parse_circuit(circuit, parallel=False, series=False,difference=False):
        """ Splits a circuit string by either dashes (series) or commas
            (parallel) outside of any paranthesis. Removes any leading 'p('
            or trailing ')' when in parallel mode or 'd('or trailing ')' when in difference mode '' """

        assert parallel != series or series != difference or difference != parallel, \
            'Exactly one of parallel or series or difference must be True'

        def count_parens(string):
            return string.count('('), string.count(')')
        
        if parallel:
            special = ','
            if circuit.endswith(')') and circuit.startswith('p('):
                circuit = circuit[2:-1]
        if difference:
            special = ','
            if circuit.endswith(')') and circuit.startswith('d('):
                circuit = circuit[2:-1]

        if series:
            special = '-'
        

        split = circuit.split(special)

        result = []
        skipped = []
        for i, sub_str in enumerate(split):
            if i not in skipped:
                if '(' not in sub_str and ')' not in sub_str:
                    result.append(sub_str)
                else:
                    open_parens, closed_parens = count_parens(sub_str)
                    if open_parens == closed_parens:
                        result.append(sub_str)
                    else:
                        uneven = True
                        while i < len(split) - 1 and uneven:
                            sub_str += special + split[i+1]

                            open_parens, closed_parens = count_parens(sub_str)
                            uneven = open_parens != closed_parens

                            i += 1
                            skipped.append(i)
                        result.append(sub_str)
        return result

    parallel = parse_circuit(circuit, parallel=True)
    series = parse_circuit(circuit, series=True)
    difference = parse_circuit(circuit, difference=True)

    if series is not None and len(series) > 1:
        eval_string += "s(["
        split = series
    elif parallel is not None and len(parallel) > 1:
        eval_string += "p(["
        split = parallel
        
    ## added for nleis.py
    elif difference is not None and len(difference) > 1:
        eval_string += "d(["
        split = difference
        
    elif series == parallel == difference :  # only single element
        split = series
        
    for i, elem in enumerate(split):
        if ',' in elem or '-' in elem :
            eval_string, index = buildCircuit(elem, frequencies,
                                              *parameters,
                                              constants=constants,
                                              eval_string=eval_string,
                                              index=index)
        else:
            param_string = ""
            raw_elem = get_element_from_name(elem)
            elem_number = check_and_eval(raw_elem).num_params
            param_list = []
            for j in range(elem_number):
                if elem_number > 1:
                    current_elem = elem + '_{}'.format(j)
                else:
                    current_elem = elem

                if current_elem in constants.keys():
                    param_list.append(constants[current_elem])
                else:
                    param_list.append(parameters[index])
                    index += 1

            param_string += str(param_list)
            new = raw_elem + '(' + param_string + ',' + str(frequencies) + ')'
            eval_string += new
            
        if i == len(split) - 1:
            if len(split) > 1:  # do not add closing brackets if single element
                eval_string += '])'
        else:
            eval_string += ','

    return eval_string, index


def calculateCircuitLength(circuit):
    """ Calculates the number of elements in the circuit.

    Parameters
    ----------
    circuit : str
        Circuit string.

    Returns
    -------
    length : int
        Length of circuit.

    """
    length = 0
    if circuit:
        extracted_elements = extract_circuit_elements(circuit)
        for elem in extracted_elements:
            raw_element = get_element_from_name(elem)
            num_params = check_and_eval(raw_element).num_params
            length += num_params
    return length

def extract_circuit_elements(circuit):
    """ Extracts circuit elements from a circuit string.

    Parameters
    ----------
    circuit : str
        Circuit string.

    Returns
    -------
    extracted_elements : list
        list of extracted elements.

    """
    p_string = [x for x in circuit if x not in 'p(),-,d()']
    extracted_elements = []
    current_element = []
    length = len(p_string)
    for i, char in enumerate(p_string):
        if char not in ints:
            current_element.append(char)
        else:
            # min to prevent looking ahead past end of list
            if p_string[min(i+1, length-1)] not in ints:
                current_element.append(char)
                extracted_elements.append(''.join(current_element))
                current_element = []
            else:
                current_element.append(char)
    extracted_elements.append(''.join(current_element))
    return extracted_elements







import string

import numpy as np

from nleis.nleis_elements_pair import (OverwriteError,
                                                circuit_elements, element, d,
                                                ElementError)
from impedance.models.circuits.circuits import CustomCircuit
from nleis.nleis import NLEISCustomCircuit

def test_porous_electrode():
   freqs = [0.001, 1.0, 1000]
   circuit_1 = CustomCircuit('TPO',initial_guess = [1,1,1])
   circuit_2 = CustomCircuit('TLM',initial_guess = [1,1,1,0,0,1000])
   Z1 = circuit_1.predict(freqs)
   Z2 = circuit_2.predict(freqs)
   assert np.allclose (Z1,Z2,atol=1e-3)



def test_RC():
   freqs = [0.001, 1.0, 1000]
   circuit_1 = CustomCircuit('RCO',initial_guess = [1,1])
   circuit_2 = CustomCircuit('p(R1,C1)',initial_guess = [1,1])
   Z1 = circuit_1.predict(freqs)
   Z2 = circuit_2.predict(freqs)

   assert np.allclose (Z1,Z2)



def test_d():
    a = np.array([5 + 6 * 1j, 2 + 3 * 1j])
    b = np.array([5 + 6 * 1j, 2 + 3 * 1j])

    answer = np.array([0, 0])
    assert np.isclose(d([a, b]), answer).all()



def test_element_function_names():
    # run a simple check to ensure there are no integers
    # in the function names
    letters = string.ascii_uppercase + string.ascii_lowercase

    for elem in circuit_elements.keys():
        print(elem)
        for char in elem:
            if elem.startswith('__'):
                continue
            assert (
                char in letters
            ), f"{char} in {elem} is not in the allowed set of {letters}"








import numpy as np
from scipy.special import iv
from impedance.models.circuits.elements import circuit_elements,ElementError,OverwriteError

def element(num_params, units, overwrite=False):
    """
    
    decorator to store metadata for a circuit element

    Parameters
    ----------
    num_params : int
        number of parameters for an element
    units : list of str
        list of units for the element parameters
    overwrite : bool (default False)
        if true, overwrites any existing element; if false,
        raises OverwriteError if element name already exists.
    
    """

    def decorator(func):
        def wrapper(p, f):
            typeChecker(p, f, func.__name__, num_params)
            return func(p, f)

        wrapper.num_params = num_params
        wrapper.units = units
        wrapper.__name__ = func.__name__
        wrapper.__doc__ = func.__doc__

        global circuit_elements
        if func.__name__ in ["s", "p"]:
            raise ElementError("cannot redefine elements 's' (series)" +
                               "or 'p' (parallel)")
        elif func.__name__ in circuit_elements and not overwrite:
            raise OverwriteError(
                f"element {func.__name__} already exists. " +
                "If you want to overwrite the existing element," +
                "use `overwrite=True`."
            )
        else:
            circuit_elements[func.__name__] = wrapper

        return wrapper

    return decorator

def d(difference):
    '''
    Subtract the second electrode from the first electrode in 2nd-NLEIS

    Notes
    -----
    
    .. math::
        

    '''

    
    z = len(difference[0])*[0 + 0*1j]
    z += difference[0]
    z += -difference[-1]
    return z
    
# manually add difference (d) operators to circuit elements w/o metadata
circuit_elements['d'] = d

@element(num_params=3, units=['Ohms', 'Ohms', 'F'])
def TPO(p, f):
    '''
    
    EIS: A macrohomogeneous porous electrode model with zero solid resistivity from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
    
    
    p0=Rpore
    p1=Rct
    p2=Cdl
    
    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.

    '''

    omega = 2*np.pi*np.array(f)

    Rpore, Rct, Cdl = p[0], p[1], p[2]

    beta = (1j*omega*Rpore*Cdl+Rpore/Rct)**(1/2)

    Z = Rpore/(beta*np.tanh(beta))
    return Z

@element(num_params=4, units=['Ohms', 'Ohms', 'F',''])
def TPOn(p, f):
    """ 
    
    2nd-NLEIS: A macrohomogeneous porous electrode model with zero solid resistivity from Ji et al. [1]

    Notes
    -----
    
    .. math::


    p0: Rpore
    p1: Rct
    p2: Cdl
    p3: ε
    
    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.

    """

    omega = 2*np.pi*np.array(f)
    Rpore, Rct, Cdl, e = p[0], p[1], p[2],p[3]
    b1 = (1j*omega*Rpore*Cdl+Rpore/Rct)**(1/2)
    b2 = (1j*2*omega*Rpore*Cdl+Rpore/Rct)**(1/2)

    f=96485.3321233100184/(8.31446261815324*298)
    sinh1 = []
    for x in b1:
        if x < 100:
            sinh1.append(np.sinh(x))
        else:
            sinh1.append(1e10)
    sinh2 = []
    cosh2 = []
    for x in b1:
        if x < 100:
            sinh2.append(np.sinh(2*x))
            cosh2.append(np.cosh(2*x))
        else:
            sinh2.append(1e10)
            cosh2.append(1e10)
    sinh3 = []
    cosh3 = []
    for x in b2:
        if x < 100:
            sinh3.append(np.sinh(x))
            cosh3.append(np.cosh(x))
        else:
            sinh3.append(1e10)
            cosh3.append(1e10)

    mf = ((Rpore**3)/Rct)*e*f/((b1*np.array(sinh1))**2)
    part1 = (b1/b2)*np.array(sinh2)/((b2**2-4*b1**2)*np.tanh(b2))
    part2 = -np.array(cosh2)/(2*(b2**2-4*b1**2))-1/(2*b2**2)
    Z = mf*(part1+part2)

    return Z

@element(num_params=5, units=['Ohms', 'Ohms', 'F','Ohms','s'])
def TDC(p, f):
    """ 
    
    EIS: A macrohomogeneous porous electrode model with cylindrical diffusion 
    and zero solid resistivity from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
    
    p0=Rpore
    p1=Rct
    p2=Cdl
    p3=Aw
    p4=τd
    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.

    """
    omega = 2*np.pi*np.array(f)
    Rpore, Rct, Cdl, Aw, τd = p[0], p[1], p[2],p[3],p[4]
    i01 = []
    i11 = []
    for x in np.sqrt(1j*omega*τd):
        if x < 100:
            i01.append(iv(0,x))
            i11.append(iv(1,x))
        else:
            i01.append(1e20)
            i11.append(1e20)
    Zd = Aw*np.array(i01)/(np.sqrt(1j*omega*τd)*np.array(i11))

    beta = (1j*omega*Rpore*Cdl+Rpore/(Zd+Rct))**(1/2)
    Z = Rpore/(beta*np.tanh(beta))
    return Z
@element(num_params=7, units=['Ohms', 'Ohms', 'F','Ohms','s','1/V',''])
def TDCn(p, f):
    """ 
    
    2nd-NLEIS: A macrohomogeneous porous electrode model with cylindrical diffusion
    and zero solid resistivity from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
        
    p0=Rpore
    p1=Rct
    p2=Cdl
    p3=Aw
    p4=τd
    p5=κ
    p6=ε
    
    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.

    """

    omega = 2*np.pi*np.array(f)
    Rpore, Rct, Cdl, Aw, τd, κ,e = p[0], p[1], p[2],p[3],p[4],p[5],p[6]

    i01 = []
    i11 = []
    for x in np.sqrt(1j*omega*τd):
        if x < 100:
            i01.append(iv(0,x))
            i11.append(iv(1,x))
        else:
            i01.append(1e20)
            i11.append(1e20)
    i02 = []
    i12 = []
    for x in np.sqrt(1j*2*omega*τd):
        if x < 100:
            i02.append(iv(0,x))
            i12.append(iv(1,x))
        else:
            i02.append(1e20)
            i12.append(1e20)
    Zd1 = Aw*np.array(i01)/(np.sqrt(1j*omega*τd)*np.array(i11))
    Zd2 = Aw*np.array(i02)/(np.sqrt(1j*2*omega*τd)*np.array(i12))

    y1 = Rct/(Zd1+Rct)
    y2 = (Zd1/(Zd1+Rct))
    
    b1 = (1j*omega*Rpore*Cdl+Rpore/(Zd1+Rct))**(1/2)
    b2 = (1j*2*omega*Rpore*Cdl+Rpore/(Zd2+Rct))**(1/2)
    
    f=96485.3321233100184/(8.31446261815324*298)
    sinh1 = []
    for x in b1:
        if x < 100:
            sinh1.append(np.sinh(x))
        else:
            sinh1.append(1e10)
    sinh2 = []
    cosh2 = []
    for x in b1:
        if x < 100:
            sinh2.append(np.sinh(2*x))
            cosh2.append(np.cosh(2*x))
        else:
            sinh2.append(1e10)
            cosh2.append(1e10)
    const= -((Rct*κ*y2**2)-Rct*e*f*y1**2)/(Zd2+Rct)
    mf = ((Rpore**3)*const/Rct)/((b1*np.array(sinh1))**2)
    part1 = (b1/b2)*np.array(sinh2)/((b2**2-4*b1**2)*np.tanh(b2))
    part2 = -np.array(cosh2)/(2*(b2**2-4*b1**2))-1/(2*b2**2)
    Z = mf*(part1+part2)

    return Z

@element(num_params=5, units=['Ohms', 'Ohms', 'F','Ohms','s'])
def TDS(p, f):
    """ 
    
    EIS: A macrohomogeneous porous electrode model with spherical diffusion
    and zero solid resistivity from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
        
    p0=Rpore
    p1=Rct
    p2=Cdl
    p3=Aw
    p4=τd
    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.
    

    """
    omega = 2*np.pi*np.array(f)
    Rpore, Rct, Cdl, Aw, τd = p[0], p[1], p[2],p[3],p[4] 

    Zd = Aw*np.tanh(np.sqrt(1j*omega*τd))/(np.sqrt(1j*omega*τd)-np.tanh(np.sqrt(1j*omega*τd)))

    beta = (1j*omega*Rpore*Cdl+Rpore/(Zd+Rct))**(1/2)
    Z = Rpore/(beta*np.tanh(beta))
    return Z
@element(num_params=7, units=['Ohms', 'Ohms', 'F','Ohms','s','1/V',''])
def TDSn(p, f):
    """ 
    
    2nd-NLEIS: A macrohomogeneous porous electrode model with spherical diffusion
    and zero solid resistivity from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
    p0=Rpore
    p1=Rct
    p2=Cdl
    p3=Aw
    p4=τd
    p5=κ
    p6=ε
    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.
    

    """

    omega = 2*np.pi*np.array(f)
    Rpore, Rct, Cdl, Aw, τd, κ,e = p[0], p[1], p[2],p[3],p[4],p[5],p[6]
    Zd1 = Aw*np.tanh(np.sqrt(1j*omega*τd))/(np.sqrt(1j*omega*τd)-np.tanh(np.sqrt(1j*omega*τd)))
    Zd2 = Aw*np.tanh(np.sqrt(1j*2*omega*τd))/(np.sqrt(1j*2*omega*τd)-np.tanh(np.sqrt(1j*2*omega*τd)))

    y1 = Rct/(Zd1+Rct)
    y2 = (Zd1/(Zd1+Rct))
    
    b1 = (1j*omega*Rpore*Cdl+Rpore/(Zd1+Rct))**(1/2)
    b2 = (1j*2*omega*Rpore*Cdl+Rpore/(Zd2+Rct))**(1/2)
    
    f=96485.3321233100184/(8.31446261815324*298)
    sinh1 = []
    for x in b1:
        if x < 100:
            sinh1.append(np.sinh(x))
        else:
            sinh1.append(1e10)
    sinh2 = []
    cosh2 = []
    for x in b1:
        if x < 100:
            sinh2.append(np.sinh(2*x))
            cosh2.append(np.cosh(2*x))
        else:
            sinh2.append(1e10)
            cosh2.append(1e10)
    const= -((Rct*κ*y2**2)-Rct*e*f*y1**2)/(Zd2+Rct)
    mf = ((Rpore**3)*const/Rct)/((b1*np.array(sinh1))**2)
    part1 = (b1/b2)*np.array(sinh2)/((b2**2-4*b1**2)*np.tanh(b2))
    part2 = -np.array(cosh2)/(2*(b2**2-4*b1**2))-1/(2*b2**2)
    Z = mf*(part1+part2)

    return Z
@element(num_params=5, units=['Ohms', 'Ohms', 'F','Ohms','s'])
def TDP(p, f):
    """ 
    
    EIS: A macrohomogeneous porous electrode model with planar diffusion 
    and zero solid resistivity from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
        
    p0 = Rpore
    p1 = Rct
    p2 = Cdl
    p3 = Aw
    p4 = τd
    
    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.
    

    """
    omega = 2*np.pi*np.array(f)
    Rpore, Rct, Cdl, Aw, τd = p[0], p[1], p[2],p[3],p[4] 

    Zd = Aw/(np.sqrt(1j*omega*τd)*np.tanh(np.sqrt(1j*omega*τd)))

    beta = (1j*omega*Rpore*Cdl+Rpore/(Zd+Rct))**(1/2)
    Z = Rpore/(beta*np.tanh(beta))

    return Z
@element(num_params=7, units=['Ohms', 'Ohms', 'F','Ohms','s','1/V',''])
def TDPn(p, f):
    """ 
    
    2nd-NLEIS: A macrohomogeneous porous electrode model with planar diffusion
    and zero solid resistivity from Ji et al. [1]
   
    Notes
    -----
    
    .. math::
        
    p0 = Rpore
    p1 = Rct
    p2 = Cdl
    p3 = Aw
    p4 = τd
    p5 = κ
    p6 = ε
    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.

    """

    omega = 2*np.pi*np.array(f)
    Rpore, Rct, Cdl, Aw, τd, κ,e = p[0], p[1], p[2],p[3],p[4],p[5],p[6]

    Zd1 = Aw/(np.sqrt(1j*omega*τd)*np.tanh(np.sqrt(1j*omega*τd)))
    Zd2 = Aw/(np.sqrt(1j*2*omega*τd)*np.tanh(np.sqrt(1j*2*omega*τd)))

    y1 = Rct/(Zd1+Rct)
    y2 = (Zd1/(Zd1+Rct))
    
    b1 = (1j*omega*Rpore*Cdl+Rpore/(Zd1+Rct))**(1/2)
    b2 = (1j*2*omega*Rpore*Cdl+Rpore/(Zd2+Rct))**(1/2)
    
    f=96485.3321233100184/(8.31446261815324*298)
    sinh1 = []
    for x in b1:
        if x < 100:
            sinh1.append(np.sinh(x))
        else:
            sinh1.append(1e10)
    sinh2 = []
    cosh2 = []
    for x in b1:
        if x < 100:
            sinh2.append(np.sinh(2*x))
            cosh2.append(np.cosh(2*x))
        else:
            sinh2.append(1e10)
            cosh2.append(1e10)
    const= -((Rct*κ*y2**2)-Rct*e*f*y1**2)/(Zd2+Rct)
    mf = ((Rpore**3)*const/Rct)/((b1*np.array(sinh1))**2)
    part1 = (b1/b2)*np.array(sinh2)/((b2**2-4*b1**2)*np.tanh(b2))
    part2 = -np.array(cosh2)/(2*(b2**2-4*b1**2))-1/(2*b2**2)
    Z = mf*(part1+part2)

    return Z
@element(num_params=2, units=['Ohm', 'F'])
def RCO(p,f):
    """ 
    
    EIS: Randles circuit 
    
    Notes
    -----
    
    .. math::
        
    p0 = Rct
    p1 = Cdl


    """
    
    w=np.array(f)*2*np.pi
    Rct=p[0]
    Cdl=p[1]
    
    Z1r=Rct/(1+(w*Rct*Cdl)**2);
    Z1i=(-w*Cdl*Rct**2)/(1+(w*Rct*Cdl)**2);
    Z1 = Z1r+1j*Z1i

    return(Z1)
@element(num_params=3, units=['Ohm', 'F',''])
def RCOn(p,f):
    '''
    
    2nd-NLEIS: Randles circuit from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
    p0 = Rct
    p1 = Cdl
    p2 = ε
    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.
    
    '''
    w=np.array(f)*2*np.pi
    Rct=p[0]
    Cdl=p[1]
    e=p[2]
    f=96485.3321233100184/(8.31446261815324*298) ## unit in 1/V
    
    Z1r=Rct/(1+(w*Rct*Cdl)**2);
    Z1i=(-w*Cdl*Rct**2)/(1+(w*Rct*Cdl)**2);
    
    tau=w*Rct*Cdl
    
    Z2r=-e*f*(Z1r**2-Z1i**2+4*tau*Z1r*Z1i)/(1+4*tau**2)
    Z2i=e*f*((Z1r**2-Z1i**2)*2*tau-2*Z1r*Z1i)/(1+4*tau**2)
    
    Z2=Z2r+1j*Z2i
    return(Z2)

@element(num_params=4, units=['Ohms', 'F','Ohms','s'])
def RCD(p,f):
    
    '''
    
    EIS: Randles circuit with planar diffusion from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
    p0 = Rct
    p1 = Cdl
    p2 = Aw
    p3 = τd
    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.
    
    '''
    omega=np.array(f)*2*np.pi
    Rct, Cdl, Aw, τd = p[0], p[1], p[2], p[3]

    Zd = Aw/(np.sqrt(1j*omega*τd)*np.tanh(np.sqrt(1j*omega*τd)))
    tau = omega*Rct*Cdl
    Z = Rct/(Rct/(Rct+Zd)+1j*tau)
    return(Z)
@element(num_params=6, units=['Ohms', 'F','Ohms','s','1/V',''])
def RCDn(p,f):
    '''
    
    2nd-NLEIS: Randles circuit with planar diffusion from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
    p0 = Rct
    p1 = Cdl
    p2 = Aw
    p3 = τd
    p4 = κ
    p5 = ε
    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.
    
    '''
    
    omega=np.array(f)*2*np.pi
    Rct, Cdl, Aw, τd, κ,e = p[0], p[1], p[2],p[3],p[4],p[5]

    Zd1 = Aw/(np.sqrt(1j*omega*τd)*np.tanh(np.sqrt(1j*omega*τd)))
    Zd2 = Aw/(np.sqrt(1j*2*omega*τd)*np.tanh(np.sqrt(1j*2*omega*τd)))

    f=96485.3321233100184/(8.31446261815324*298)


    tau = omega*Rct*Cdl
    y1 = Rct/(Zd1+Rct)
    y2 = (Zd1/(Zd1+Rct))
    
    Z1 = Rct/(y1+1j*tau)
    const= ((Rct*κ*y2**2)-Rct*e*f*y1**2)/(Zd2+Rct)

    Z2 = (const*Z1**2)/(2*tau*1j+Rct/(Zd2+Rct))
    
    return(Z2)
@element(num_params=4, units=['Ohms', 'F','Ohms','s'])
def RCS(p,f):
    '''
    
    EIS: Randles circuit with spherical diffusion from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
    p0 = Rct
    p1 = Cdl
    p2 = Aw
    p3 = τd
    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.
    
    '''
    omega=np.array(f)*2*np.pi
    Rct, Cdl, Aw, τd = p[0], p[1], p[2], p[3]

    Zd = Aw*np.tanh(np.sqrt(1j*omega*τd))/(np.sqrt(1j*omega*τd)-np.tanh(np.sqrt(1j*omega*τd)))

    tau = omega*Rct*Cdl
    Z = Rct/(Rct/(Rct+Zd)+1j*tau)
    return(Z)
@element(num_params=6, units=['Ohms', 'F','Ohms','s','1/V',''])
def RCSn(p,f):
    '''
    
    2nd-NLEIS: Randles circuit with spherical diffusion from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
    p0 = Rct
    p1 = Cdl
    p2 = Aw
    p3 = τd
    p4 = κ
    p5 = ε
    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.
    
    '''    
    

    omega=np.array(f)*2*np.pi
    Rct, Cdl, Aw, τd, κ,e = p[0], p[1], p[2],p[3],p[4],p[5]

    Zd1 = Aw*np.tanh(np.sqrt(1j*omega*τd))/(np.sqrt(1j*omega*τd)-np.tanh(np.sqrt(1j*omega*τd)))
    Zd2 = Aw*np.tanh(np.sqrt(1j*2*omega*τd))/(np.sqrt(1j*2*omega*τd)-np.tanh(np.sqrt(1j*2*omega*τd)))
    
    f=96485.3321233100184/(8.31446261815324*298)


    tau = omega*Rct*Cdl
    y1 = Rct/(Zd1+Rct)
    y2 = (Zd1/(Zd1+Rct))
    
    Z1 = Rct/(y1+1j*tau)
    const= ((Rct*κ*y2**2)-Rct*e*f*y1**2)/(Zd2+Rct)

    Z2 = (const*Z1**2)/(2*tau*1j+Rct/(Zd2+Rct))
    
    return(Z2)



@element(num_params=6, units=['Ohm', 'Ohm', 'F', 'Ohm', 'F', ''])
def TLM(p,f):
    
    """ 
    
    EIS： General discrete transmission line model built  Randles circuit
    
    Notes
    -----
    
    .. math::
        
    p0: Rpore
    p1: Rct,bulk
    p2: Cdl,bulk
    p3: Rct,surface
    p4: Cdl,surface
    p5: N (number of circuit element)


    """
    
    
    N=int(p[5])
    frequencies = np.array(f)

    Rct=p[1]*N
    Cdl=p[2]/N
    Rpore=p[0]/N
    Rs=p[3]*N
    Cs=p[4]/N

    Z1b = RCO([Rct,Cdl],frequencies)
    Z1s = RCO([Rs,Cs],frequencies)
    Zran = Z1b + Z1s
    Req = Zran
    for i in range(1,N):
        
        Req_inv=(1/(Req+Rpore))+1/Zran
        Req=1/Req_inv
        
    return (Req)
###
@element(num_params=8, units=['Ohm', 'Ohm', 'F','','Ohm', 'F','', ''])
def TLMn(p,f):
    """ 
    
    2nd-NLEIS: Second harmonic nonlinear discrete transmission line model 
    built based on the nonlinear Randles circuit from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
    p0: Rpore
    p1: Rct,bulk
    p2: Cdl,bulk

    p3: Rct,surface
    p4: Cdl,surface
    p5: N (number of circuit element)
    p6: ε,bulk
    p7: ε,surface
   
    
    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.

    """
    I=mTi(p[0:6],f) # calculate the current fraction (1st harmonic)

    N=int(p[5])
    frequencies = np.array(f)
    
    Rpore=p[0]/N
    Rct=p[1]*N
    Cdl=p[2]/N
    Rs=p[3]*N
    Cs=p[4]/N
    
    eb=p[6]
    es=p[7]
    f=96485.3321233100184/(8.31446261815324*298) ## unit in 1/V
    Z1b = RCO([Rct,Cdl],frequencies)
    Z1s = RCO([Rs,Cs],frequencies)
    Z2b = RCOn([Rct,Cdl,eb],frequencies)
    Z2s = RCOn([Rs,Cs,es],frequencies)
    Z1b2t = RCO([Rct,Cdl],2*frequencies)
    Z1s2t = RCO([Rs,Cs],2*frequencies)
    Z1 = Z1b+Z1s
    Z2 = Z2b+Z2s
    Z12t = Z1b2t+Z1s2t
    if N==1:
        return(Z2)
    
    if N==2:
        sum1=Z1**2/(2*Z1+Rpore)**2
        sum2=(Z12t*Rpore+Rpore**2)/((2*Z12t+Rpore)*(2*Z1+Rpore))
        Z=(sum1+sum2)*Z2
        return(Z)
    Z=np.zeros((len(frequencies)),dtype = complex)
    for freq in range(0,len(frequencies)):
        Ii=I[freq]
        
        A = np.arange(N-1,0,-1)
        A1 = np.arange(N-1,0,-1)
        
        for i in range (0,N-2):
            for j in range(0,N-1-i):
                A1[j]=A1[j]-1
            A=np.vstack((A,A1))
        A=A*Rpore 
        A=np.append(A, np.zeros((N-1,1)), axis = 1)
        A=np.append(A, np.zeros((1,N)), axis = 0)
        A2=np.zeros((N-1,N))
        for i in range(0,N-1):
            A2[i,0]+=1
            A2[i,N-1-i]-=1
        A2=np.vstack((A2,np.zeros(N)))
        A2=A2*Z12t[freq]
        
        A3=np.vstack((np.zeros((N-1,N)),np.ones(N)))
        
        Ax = A2+A+A3
        
        b=np.zeros((N,1),dtype = complex)

        for i in range (0,N-1):
            b[i]=Ii[-1]**2-Ii[i]**2
        
        I2=np.linalg.solve(Ax,-b*Z2[freq])
        Z[freq]=Z2[freq]*Ii[0]**2+I2[-1]*Z12t[freq]
    return(Z)

@element(num_params=6, units=['Ohm', 'Ohm', 'F', 'Ohm', 'F', ''])
def mTi(p,f):
    
    """ 
    
    EIS: current distribution of  discrete transmission line model 
    built based on the  Randles circuit from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
    p0: Rpore
    p1: Rct,bulk
    p2: Cdl,bulk
    p3: Rct,surface
    p4: Cdl,surface
    p5: N (number of circuit element)

    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.

    """
    
    
    N=int(p[5])
    frequencies = np.array(f)

    Rct=p[1]*N
    Cdl=p[2]/N
    Rpore=p[0]/N
    Rs=p[3]*N
    Cs=p[4]/N
    Z1b = RCO([Rct,Cdl],frequencies)
    Z1s = RCO([Rs,Cs],frequencies)
    Zran = Z1b + Z1s
    Req = Zran
    for i in range(1,N):
        
        Req_inv=(1/(Req+Rpore))+1/Zran
        Req=1/Req_inv
        
    Req=Req+Rpore
    
    I=np.zeros((len(frequencies),N),dtype=complex)
    for freq in range (0,len(frequencies)):
        b1=np.ones(N)*Req[freq]

        A=np.identity(N)*Zran[freq]
        
        A1=np.ones((N,N))*Rpore
        
        for i in range(0,N):
            
            A1[i,:]=A1[i,:]*(i+1)
            
            for j in range(0,i):
                
                A[i][j]=-(i-j)*Rpore

        A = A+A1
        
        b = b1
        
        I[freq,:] = np.linalg.solve(A, b)   
    return(I)



@element(num_params=8, units=['Ohm', 'Ohm', 'F','Ohm', 's', 'Ohm', 'F', ''])
def TLMS(p,f):
    
    """ 
    
    EIS: General discrete transmission line model 
    built based on the Randles circuit with spherical diffusion from Ji et al.[1]
    
    Notes
    -----
    
    .. math::
    
    p0: Rpore
    p1: Rct,bulk
    p2: Cdl,bulk
    p3: Aw,bulk
    p4: τ,bulk
    p5: Rct,surface
    p6: Cdl,surface
    p7: N (number of circuit element)

    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.

    """
    
    N=int(p[7])
    frequencies = np.array(f)

    Rpore=p[0]/N
    Rct=p[1]*N
    Cdl=p[2]/N
    Aw=p[3]*N
    τd=p[4]
    Rs=p[5]*N
    Cs=p[6]/N

    Z1b = RCS([Rct,Cdl,Aw,τd],frequencies)

    Z1s = RCO([Rs,Cs],frequencies)
    Zran = Z1b + Z1s
    Req = Zran
    for i in range(1,N):
        
        Req_inv=(1/(Req+Rpore))+1/Zran
        Req=1/Req_inv
        
    return (Req)
    
@element(num_params=11, units=['Ohm', 'Ohm', 'F','Ohm', 's','Ohm', 'F','','1/V','', ''])
def TLMSn(p,f):
    """ 
    
    2nd-NLEIS: Second harmonic nonlinear discrete transmission line model 
    built based on the Randles circuit with spherical diffusion from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
    p0: Rpore
    p1: Rct,bulk
    p2: Cdl,bulk
    p3: Aw,bulk
    p4: τ,bulk
    p5: Rct,surface
    p6: Cdl,surface
    p7: N (number of circuit element)
    p8: κ,bulk
    p9: ε,bulk
    p10: ε,surface

        

    """
    I=mTiS(p[0:8],f) # calculate the current fraction (1st harmonic)

    N=int(p[7])
    
    frequencies = np.array(f)

    Rpore=p[0]/N
    Rct=p[1]*N
    Cdl=p[2]/N
    Aw=p[3]*N
    τd=p[4]
    Rs=p[5]*N
    Cs=p[6]/N
    κ = p[8]
    eb = p[9]
    es = p[10]

    Z1b = RCS([Rct,Cdl,Aw,τd],frequencies)
    Z1s = RCO([Rs,Cs],frequencies)
    Z2b = RCSn([Rct,Cdl,Aw,τd,κ,eb],frequencies)
    Z2s = RCOn([Rs,Cs,es],frequencies)
    Z1b2t = RCS([Rct,Cdl,Aw,τd],2*frequencies)
    Z1s2t = RCO([Rs,Cs],2*frequencies)
    
    Z1 = Z1b+Z1s
    Z2 = Z2b+Z2s
    Z12t = Z1b2t+Z1s2t

    if N==1:
        return(Z2)
    
    if N==2:
        sum1=Z1**2/(2*Z1+Rpore)**2
        sum2=(Z12t*Rpore+Rpore**2)/((2*Z12t+Rpore)*(2*Z1+Rpore))
        Z=(sum1+sum2)*Z2
        return(Z)
        
    Z=np.zeros(len(frequencies),dtype = complex)
    for freq in range(0,len(frequencies)):
        Ii=I[freq]
        
        A = np.arange(N-1,0,-1)
        A1 = np.arange(N-1,0,-1)
        
        for i in range (0,N-2):
            for j in range(0,N-1-i):
                A1[j]=A1[j]-1
            A=np.vstack((A,A1))
        A=A*Rpore 
        A=np.append(A, np.zeros((N-1,1)), axis = 1)
        A=np.append(A, np.zeros((1,N)), axis = 0)
        A2=np.zeros((N-1,N))
        for i in range(0,N-1):
            A2[i,0]+=1
            A2[i,N-1-i]-=1
        A2=np.vstack((A2,np.zeros(N)))
        A2=A2*Z12t[freq]
        
        A3=np.vstack((np.zeros((N-1,N)),np.ones(N)))
        
        Ax = A2+A+A3
        
        b=np.zeros((N,1),dtype = complex)

        for i in range (0,N-1):
            b[i]=Ii[-1]**2-Ii[i]**2
        
        I2=np.linalg.solve(Ax,-b*Z2[freq])
        Z[freq]=Z2[freq]*Ii[0]**2+I2[-1]*Z12t[freq]
        # Z[freq]=Z2[freq]*Ii[0]**2

    return(Z)
@element(num_params=8, units=['Ohm', 'Ohm', 'F','Ohm', 's', 'Ohm', 'F', ''])
def mTiS(p,f):
    
    """ 
    
    EIS: current distribution of nonlinear discrete transmission line model 
    built based on the Randles circuit with spherical diffusion from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
    p0: Rpore
    p1: Rct,bulk
    p2: Cdl,bulk
    p3: Aw,bulk
    p4: τ,bulk
    p5: Rct,surface
    p6: Cdl,surface
    p7: N (number of circuit element)

    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.

    """
    N=int(p[7])
    frequencies = np.array(f)

    Rpore=p[0]/N
    Rct=p[1]*N
    Cdl=p[2]/N
    Aw=p[3]*N
    τd=p[4]
    Rs=p[5]*N
    Cs=p[6]/N

    Z1b = RCS([Rct,Cdl,Aw,τd],frequencies)

    Z1s = RCO([Rs,Cs],frequencies)
    Zran = Z1b + Z1s
    Req = Zran
    for i in range(1,N):
        
        Req_inv=(1/(Req+Rpore))+1/Zran
        Req=1/Req_inv
        
    Req=Req+Rpore
    
    I=np.zeros((len(frequencies),N),dtype=complex)
    for freq in range (0,len(frequencies)):
        b1=np.ones(N)*Req[freq]

        A=np.identity(N)*Zran[freq]
        
        A1=np.ones((N,N))*Rpore
        
        for i in range(0,N):
            
            A1[i,:]=A1[i,:]*(i+1)
            
            for j in range(0,i):
                
                A[i][j]=-(i-j)*Rpore

        A = A+A1
        
        b = b1
        
        I[freq,:] = np.linalg.solve(A, b)   

    return(I)
    
@element(num_params=11, units=['Ohm', 'Ohm', 'F','Ohm', 's','Ohm', 'F','','1/V','', ''])
def mTiSn(p,f):
    """ 
    
    2nd-NLEIS: nonlinear current distribution of nonlinear discrete transmission line model
      built based on the Randles circuit with spherical diffusion from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
    p0: Rpore
    p1: Rct,bulk
    p2: Cdl,bulk
    p3: Aw,bulk
    p4: τ,bulk
    p5: Rct,surface
    p6: Cdl,surface
    p7: N (number of circuit element)
    p8: κ,bulk
    p9: ε,bulk
    p10: ε,surface

        

    """
    I=mTiS(p[0:8],f) # calculate the current fraction (1st harmonic)

    N=int(p[7])
    
    frequencies = np.array(f)

    Rpore=p[0]/N
    Rct=p[1]*N
    Cdl=p[2]/N
    Aw=p[3]*N
    τd=p[4]
    Rs=p[5]*N
    Cs=p[6]/N
    κ = p[8]
    eb = p[9]
    es = p[10]

    Z1b = RCS([Rct,Cdl,Aw,τd],frequencies)
    Z1s = RCO([Rs,Cs],frequencies)
    Z2b = RCSn([Rct,Cdl,Aw,τd,κ,eb],frequencies)
    Z2s = RCOn([Rs,Cs,es],frequencies)
    Z1b2t = RCS([Rct,Cdl,Aw,τd],2*frequencies)
    Z1s2t = RCO([Rs,Cs],2*frequencies)
    
    Z1 = Z1b+Z1s
    Z2 = Z2b+Z2s
    Z12t = Z1b2t+Z1s2t

    if N==1:
        return(Z2)
    
    if N==2:
        sum1=Z1**2/(2*Z1+Rpore)**2
        sum2=(Z12t*Rpore+Rpore**2)/((2*Z12t+Rpore)*(2*Z1+Rpore))
        Z=(sum1+sum2)*Z2
        return(Z)
        
    I2=np.zeros((len(frequencies),N),dtype=complex)

    for freq in range(0,len(frequencies)):
        Ii=I[freq]
        
        A = np.arange(N-1,0,-1)
        A1 = np.arange(N-1,0,-1)
        
        for i in range (0,N-2):
            for j in range(0,N-1-i):
                A1[j]=A1[j]-1
            A=np.vstack((A,A1))
        A=A*Rpore 
        A=np.append(A, np.zeros((N-1,1)), axis = 1)
        A=np.append(A, np.zeros((1,N)), axis = 0)
        A2=np.zeros((N-1,N))
        for i in range(0,N-1):
            A2[i,0]+=1
            A2[i,N-1-i]-=1
        A2=np.vstack((A2,np.zeros(N)))
        A2=A2*Z12t[freq]
        
        A3=np.vstack((np.zeros((N-1,N)),np.ones(N)))
        
        Ax = A2+A+A3
        
        b=np.zeros((N,1),dtype = complex)

        for i in range (0,N-1):
            b[i]=Ii[-1]**2-Ii[i]**2
        
        I2[freq,:]=np.linalg.solve(Ax,-b*Z2[freq]).flatten()[::-1] ## reverse the order to display the correct result from small to larger N

        

    return(I2)



@element(num_params=8, units=['Ohm', 'Ohm', 'F','Ohm', 's', 'Ohm', 'F', ''])
def TLMD(p,f):
    
    """ 
    
    EIS: general discrete transmission line model 
    built based on the Randles circuit with planar diffusion from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
    
    p0: Rpore
    p1: Rct,bulk
    p2: Cdl,bulk
    p3: Aw,bulk
    p4: τ,bulk
    p5: Rct,surface
    p6: Cdl,surface
    p7: N (number of circuit element)

    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.

    """
    
    N=int(p[7])
    frequencies = np.array(f)

    Rpore=p[0]/N
    Rct=p[1]*N
    Cdl=p[2]/N
    Aw=p[3]*N
    τd=p[4]
    Rs=p[5]*N
    Cs=p[6]/N

    Z1b = RCD([Rct,Cdl,Aw,τd],frequencies)

    Z1s = RCO([Rs,Cs],frequencies)
    Zran = Z1b + Z1s
    Req = Zran
    for i in range(1,N):
        
        Req_inv=(1/(Req+Rpore))+1/Zran
        Req=1/Req_inv
    return (Req)
    
@element(num_params=11, units=['Ohm', 'Ohm', 'F','Ohm', 's','Ohm', 'F','','1/V','', ''])
def TLMDn(p,f):
    """ 
    
    2nd-NLEIS: Second harmonic nonlinear discrete transmission line model 
    built based on the Randles circuit with planar diffusion from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
    p0: Rpore
    p1: Rct,bulk
    p2: Cdl,bulk
    p3: Aw,bulk
    p4: τ,bulk
    p5: Rct,surface
    p6: Cdl,surface
    p7: N (number of circuit element)
    p8: κ,bulk
    p9: ε,bulk
    p10: ε,surface

        

    """

    I=mTiD(p[0:8],f) # calculate the current fraction (1st harmonic)

    N=int(p[7])
    frequencies = np.array(f)
    
    Rpore=p[0]/N
    Rct=p[1]*N
    Cdl=p[2]/N
    Aw=p[3]*N
    τd=p[4]
    Rs=p[5]*N
    Cs=p[6]/N
    κ = p[8]
    eb = p[9]
    es = p[10]

    f=96485.3321233100184/(8.31446261815324*298) ## unit in 1/V


    Z1b = RCD([Rct,Cdl,Aw,τd],frequencies)
    Z1s = RCO([Rs,Cs],frequencies)
    Z2b = RCDn([Rct,Cdl,Aw,τd,κ,eb],frequencies)
    Z2s = RCOn([Rs,Cs,es],frequencies)
    Z1b2t = RCD([Rct,Cdl,Aw,τd],2*frequencies)
    Z1s2t = RCO([Rs,Cs],2*frequencies)

    Z1 = Z1b+Z1s
    Z2 = Z2b+Z2s
    Z12t = Z1b2t+Z1s2t

    if N==1:
        return(Z2)
    
    if N==2:
        sum1=Z1**2/(2*Z1+Rpore)**2
        sum2=(Z12t*Rpore+Rpore**2)/((2*Z12t+Rpore)*(2*Z1+Rpore))
        Z=(sum1+sum2)*Z2
        return(Z)
    Z=np.zeros((len(frequencies)),dtype = complex)
    for freq in range(0,len(frequencies)):
        Ii=I[freq]
        
        A = np.arange(N-1,0,-1)
        A1 = np.arange(N-1,0,-1)
        
        for i in range (0,N-2):
            for j in range(0,N-1-i):
                A1[j]=A1[j]-1
            A=np.vstack((A,A1))
        A=A*Rpore 
        A=np.append(A, np.zeros((N-1,1)), axis = 1)
        A=np.append(A, np.zeros((1,N)), axis = 0)
        A2=np.zeros((N-1,N))
        for i in range(0,N-1):
            A2[i,0]+=1
            A2[i,N-1-i]-=1
        A2=np.vstack((A2,np.zeros(N)))
        A2=A2*Z12t[freq]
        
        A3=np.vstack((np.zeros((N-1,N)),np.ones(N)))
        
        Ax = A2+A+A3
        
        b=np.zeros((N,1),dtype = complex)

        for i in range (0,N-1):
            b[i]=Ii[-1]**2-Ii[i]**2
        
        I2=np.linalg.solve(Ax,-b*Z2[freq])
        Z[freq]=Z2[freq]*Ii[0]**2+I2[-1]*Z12t[freq]
    return(Z)

@element(num_params=8, units=['Ohm', 'Ohm', 'F','Ohm', 's', 'Ohm', 'F', ''])
def mTiD(p,f):
    
    """ 
    
    EIS: current distribution of discrete transmission line model 
    built based on the Randles circuit with planar diffusion from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
    p0: Rpore
    p1: Rct,bulk
    p2: Cdl,bulk
    p3: Aw,bulk
    p4: τ,bulk
    p5: Rct,surface
    p6: Cdl,surface
    p7: N (number of circuit element)

    
    [1] Y. Ji, D.T. Schwartz, 
    Second-Harmonic Nonlinear Electrochemical Impedance Spectroscopy: 
    I. Analytical theory and equivalent circuit representations for planar and porous electrodes. 
    J. Electrochem. Soc. (2023). `doi: 10.1149/1945-7111/ad15ca
    <https://doi.org/10.1149/1945-7111/ad15ca>`_.

    """

    N=int(p[7])
    frequencies = np.array(f)


    Rpore=p[0]/N
    Rct=p[1]*N
    Cdl=p[2]/N
    Aw=p[3]*N
    τd=p[4]
    Rs=p[5]*N
    Cs=p[6]/N

    Z1b = RCD([Rct,Cdl,Aw,τd],frequencies)

    Z1s = RCO([Rs,Cs],frequencies)
    Zran = Z1b + Z1s
    Req = Zran
    for i in range(1,N):
        
        Req_inv=(1/(Req+Rpore))+1/Zran
        Req=1/Req_inv
        
    Req=Req+Rpore
    
    I=np.zeros((len(frequencies),N),dtype=complex)
    for freq in range (0,len(frequencies)):
        b1=np.ones(N)*Req[freq]

        A=np.identity(N)*Zran[freq]
        
        A1=np.ones((N,N))*Rpore
        
        for i in range(0,N):
            
            A1[i,:]=A1[i,:]*(i+1)
            
            for j in range(0,i):
                
                A[i][j]=-(i-j)*Rpore

        A = A+A1
        
        b = b1
        
        I[freq,:] = np.linalg.solve(A, b)   
    return(I)

@element(num_params=11, units=['Ohm', 'Ohm', 'F','Ohm', 's','Ohm', 'F','','1/V','', ''])
def mTiDn(p,f):
    """ 
    
    2nd-NLEIS: current distribution of Second harmonic nonlinear discrete transmission line model 
    built based on the Randles circuit with planar diffusion from Ji et al. [1]
    
    Notes
    -----
    
    .. math::
        
    p0: Rpore
    p1: Rct,bulk
    p2: Cdl,bulk
    p3: Aw,bulk
    p4: τ,bulk
    p5: Rct,surface
    p6: Cdl,surface
    p7: N (number of circuit element)
    p8: κ,bulk
    p9: ε,bulk
    p10: ε,surface

        

    """

    I=mTiD(p[0:8],f) # calculate the current fraction (1st harmonic)

    N=int(p[7])
    frequencies = np.array(f)
    
    Rpore=p[0]/N
    Rct=p[1]*N
    Cdl=p[2]/N
    Aw=p[3]*N
    τd=p[4]
    Rs=p[5]*N
    Cs=p[6]/N
    κ = p[8]
    eb = p[9]
    es = p[10]

    f=96485.3321233100184/(8.31446261815324*298) ## unit in 1/V


    Z1b = RCD([Rct,Cdl,Aw,τd],frequencies)
    Z1s = RCO([Rs,Cs],frequencies)
    Z2b = RCDn([Rct,Cdl,Aw,τd,κ,eb],frequencies)
    Z2s = RCOn([Rs,Cs,es],frequencies)
    Z1b2t = RCD([Rct,Cdl,Aw,τd],2*frequencies)
    Z1s2t = RCO([Rs,Cs],2*frequencies)

    Z1 = Z1b+Z1s
    Z2 = Z2b+Z2s
    Z12t = Z1b2t+Z1s2t

    if N==1:
        return(Z2)
    
    if N==2:
        sum1=Z1**2/(2*Z1+Rpore)**2
        sum2=(Z12t*Rpore+Rpore**2)/((2*Z12t+Rpore)*(2*Z1+Rpore))
        Z=(sum1+sum2)*Z2
        return(Z)

    I2=np.zeros((len(frequencies),N),dtype=complex)

    for freq in range(0,len(frequencies)):
        Ii=I[freq]
        
        A = np.arange(N-1,0,-1)
        A1 = np.arange(N-1,0,-1)
        
        for i in range (0,N-2):
            for j in range(0,N-1-i):
                A1[j]=A1[j]-1
            A=np.vstack((A,A1))
        A=A*Rpore 
        A=np.append(A, np.zeros((N-1,1)), axis = 1)
        A=np.append(A, np.zeros((1,N)), axis = 0)
        A2=np.zeros((N-1,N))
        for i in range(0,N-1):
            A2[i,0]+=1
            A2[i,N-1-i]-=1
        A2=np.vstack((A2,np.zeros(N)))
        A2=A2*Z12t[freq]
        
        A3=np.vstack((np.zeros((N-1,N)),np.ones(N)))
        
        Ax = A2+A+A3
        
        b=np.zeros((N,1),dtype = complex)

        for i in range (0,N-1):
            b[i]=Ii[-1]**2-Ii[i]**2
        
        I2[freq,:]=np.linalg.solve(Ax,-b*Z2[freq]).flatten()[::-1] ## reverse the order to display the correct result from small to larger N
        
    return(Z)


def get_element_from_name(name):
    excluded_chars = '0123456789_'
    return ''.join(char for char in name if char not in excluded_chars)


def typeChecker(p, f, name, length):
    assert isinstance(p, list), \
        'in {}, input must be of type list'.format(name)
    for i in p:
        assert isinstance(i, (float, int, np.int32, np.float64)), \
            'in {}, value {} in {} is not a number'.format(name, i, p)
    for i in f:
        assert isinstance(i, (float, int, np.int32, np.float64)), \
            'in {}, value {} in {} is not a number'.format(name, i, f)
    assert len(p) == length, \
        'in {}, input list must be length {}'.format(name, length)
    return



from nleis import __version__
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nleis",
    version= __version__ ,
    author="Yuefan Ji",
    author_email="yuefan@uw.edu",
    description="A NLEIS toolbox for impedance.py that provides RC level nonlinear equivalent modeling and analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://nleispy.readthedocs.io/en/latest/",
    project_urls={
        "Bug Tracker": "https://github.com/yuefan98/nleis.py",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(include = ['nleis']),
    package_data={
        "nleis": ["data/*",'nleis_tests/*'],

	},
    
    python_requires=">=3.6",
    install_requires=['altair>=3.0', 'matplotlib>=3.5',
                  'numpy>=1.14', 'scipy>=1.0','impedance>=1.7.1'],
)
